

from hgc.constants import read_write
import pytest
import hgc


def test_csv_to_named_tuple():
    """  test whether csv is read correctly """
    atoms, ions, properties = read_write.convert_csv_to_tuples()
    assert atoms['H'].feature == 'H'
    assert atoms['H'].name == 'Hydrogen'
    assert atoms['H'].unit == 'mg/L'
    assert atoms['H'].mw == 1.00794
    assert atoms['H'].oxidized == 1
    assert atoms['H'].reduced == 0

    assert ions['CH4'].feature == 'CH4'
    assert ions['CH4'].name == 'methane as mg CH4/L'
    assert ions['CH4'].example == 'read'
    assert ions['CH4'].unit == 'mg/L'
    assert ions['CH4'].valence == 0
    assert ions['CH4'].mw == 16.04246
    assert ions['PO4_ortho'].mw == 'calculate'
    assert ions['doc'].mw == 'unknown'

    assert properties['ec_lab'].feature == 'ec_lab'
    assert properties['ec_lab'].name == 'EC converted to 20°C in lab'
    assert properties['ec_lab'].example == 'read'
    assert properties['ec_lab'].unit == 'μS/cm'


def test_get_mw():
    """ assert correct molar weights are returned """
    assert hgc.mw('Hg') == 200.59
    assert hgc.mw('Fe') == 55.845
    with pytest.raises(KeyError):
        hgc.mw('NH4')


def test_get_units():
    """ assert correct units are returned """
    assert hgc.units('Hg') == 'μg/L'
    assert hgc.units('Fe') == 'mg/L'
    assert hgc.units('NH4') == 'mg/L'
    assert hgc.units('SO4') == 'mg/L as SO4'
    assert hgc.units('alkalinity') == 'mg/L as HCO3'
    assert hgc.units('eh_field') == 'mV'
    assert hgc.units('ph') == '-'

    with pytest.raises(KeyError):
        hgc.units('pH')
    with pytest.raises(KeyError):
        hgc.units('some other non sense')


''' Testing of the integration of phreeqpython in hgc '''
import logging

import numpy as np
import pandas as pd
import pytest
from phreeqpython import PhreeqPython, Solution

import hgc
from hgc.constants.constants import mw
from . import test_directory



@pytest.fixture(name='mineral_data')
def fixture_mineral_data():
    ''' fixture that loads the test data into a dataframe and makes it valid
        (if possible). the dataframe is returned '''
    df = pd.read_csv(test_directory / 'data' / 'dataset_basic.csv',
                     skiprows=[1], index_col=None)
    df[df.hgc.hgc_cols] = df[df.hgc.hgc_cols].astype(float)
    df.hgc.make_valid()
    return df


@pytest.fixture(name='consolidated_data')
def fixture_consolidated_data():
    ''' same as fixture mineral_data, but now
        including consolidate before returning
        df '''
    df = pd.read_csv(test_directory / 'data' / 'dataset_basic.csv',
                     skiprows=[1], index_col=None)
    df[df.hgc.hgc_cols] = df[df.hgc.hgc_cols].astype(float)
    df.hgc.make_valid()
    df.hgc.consolidate(inplace=True, use_so4=None, use_ph='lab')
    return df


@pytest.fixture(name='phreeqpython_solutions_excel')
def fixture_phreeqpython_solutions_excel():
    ''' Add the solutions of the excel file manually to test
        whether they are added correctly and to easily check
        whether all derived quantities like EC, and SI of calicite
        are correct by hgc methods '''
        #     {'temp': 10, 'Alkalinity': 0, 'O(0)': '11 ', 'C(-4)': '0 as CH4', 'pH': 4.5, 'Na': '9 charge', 'K': 0.4, 'Ca': 1, 'Mg': 1.1, 'Fe': 0.1, 'Mn': 0.02, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'},
        #     {'temp': 10, 'Alkalinity': 0, 'O(0)': '2 ', 'C(-4)': '0 as CH4', 'pH': 4.3, 'Na': '15 charge', 'K': 1.1, 'Ca': 2, 'Mg': 1.6, 'Fe': 0.29, 'Mn': 0.05, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10, 'Alkalinity': 0, 'O(0)': '1 ', 'C(-4)': '0 as CH4', 'pH': 4.4, 'Na': '19 charge', 'K': 1.8, 'Ca': 2, 'Mg': 3.0, 'Fe': 0.12, 'Mn': 0.06, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10, 'Alkalinity': 7, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 5.5, 'Na': '20 charge', 'K': 2.1, 'Ca': 3, 'Mg': 3.9, 'Fe': 2.33, 'Mn': 0.13, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10, 'Alkalinity': 13, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 5.7, 'Na': '21 charge', 'K': 2.8, 'Ca': 5, 'Mg': 5.4, 'Fe': 3.43, 'Mn': 0.03, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10, 'Alkalinity': 23, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 6.2, 'Na': '24 charge', 'K': 1.8, 'Ca': 13, 'Mg': 2.7, 'Fe': 0.7, 'Mn': 0.09, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10, 'Alkalinity': 92, 'O(0)': '0 ', 'C(-4)': '0.14 as CH4', 'pH': 7.4, 'Na': '29 charge', 'K': 1.9, 'Ca': 38, 'Mg': 2.8, 'Fe': 0.3, 'Mn': 0.13, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.1, 'Alkalinity': 146, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 7.7, 'Na': '29 charge', 'K': 2.0, 'Ca': 51, 'Mg': 3.4, 'Fe': 0.51, 'Mn': 0.13, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.2, 'Alkalinity': 151, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 7.7, 'Na': '15 charge', 'K': 1.4, 'Ca': 48, 'Mg': 3.1, 'Fe': 3.4, 'Mn': 0.18, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.5, 'Alkalinity': 156, 'O(0)': '0 ', 'C(-4)': '5 as CH4', 'pH': 7.6, 'Na': '15 charge', 'K': 1.2, 'Ca': 48, 'Mg': 2.5, 'Fe': 1.22, 'Mn': 0.13, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.7, 'Alkalinity': 169, 'O(0)': '0 ', 'C(-4)': '9.1 as CH4', 'pH': 8.1, 'Na': '24 charge', 'K': 9.5, 'Ca': 13, 'Mg': 21, 'Fe': 0.69, 'Mn': 0.04, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.8, 'Alkalinity': 398, 'O(0)': '0 ', 'C(-4)': '18.2 as CH4', 'pH': 8.5, 'Na': '155 charge', 'K': 11.5, 'Ca': 3, 'Mg': 3.1, 'Fe': 0.62, 'Mn': 0.04, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.7, 'Alkalinity': 525, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 8.2, 'Na': '190 charge', 'K': 16.8, 'Ca': 9, 'Mg': 9.5, 'Fe': 0.17, 'Mn': 0.12, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.8, 'Alkalinity': 801, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 7.7, 'Na': '445 charge', 'K': 58, 'Ca': 27, 'Mg': 19.9, 'Fe': 1.6, 'Mn': 0.12, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 10.8, 'Alkalinity': 903, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 8.5, 'Na': '910 charge', 'K': 134, 'Ca': 66, 'Mg': 71, 'Fe': 5.4, 'Mn': 0.3, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        #     {'temp': 11.0, 'Alkalinity': 2635, 'O(0)': '0 ', 'C(-4)': '0 as CH4', 'pH': 7.8, 'Na': '3380 charge', 'K': 95, 'Ca': 160, 'Mg': 390, 'Fe': 2.3, 'Mn': 0.03, 'Amm': '1.29', 'Si': '0.2 as SiO2', 'Cl': 16, 'S(6)': '7 as SO4', 'N(5)': '3.4 as NO3', 'F': 0.027, 'Mn(7)': '5.2 as KMnO4'}
        # ]
    pp = PhreeqPython()
    solution_dictionaries = [
        {'temp': 10,   'Alkalinity':    '0 as HCO3', 'O(0)': '11 ', 'C(-4)':    '0 as CH4', 'pH': 4.5, 'Na':    '9', 'K':   0.4, 'Ca':   1, 'Mg':   1.1, 'Fe': 0.10, 'Mn': 0.02, 'Amm':  1.29, 'Si':  '0.2 as SiO2', 'Cl':   16, 'S(6)':  '7 as SO4', 'N(5)':  '3.4 as NO3', 'F':  0.02700, 'Mn(7)':   '5.2 as KMnO4'},
        {'temp': 10,   'Alkalinity':    '0 as HCO3', 'O(0)':  '2 ', 'C(-4)':    '0 as CH4', 'pH': 4.3, 'Na':   '15', 'K':   1.1, 'Ca':   2, 'Mg':   1.6, 'Fe': 0.29, 'Mn': 0.05, 'Amm':  0.05, 'Si': '11.2 as SiO2', 'Cl':   19, 'S(6)': '16 as SO4', 'N(5)': '22.6 as NO3', 'F':  0.08000, 'Mn(7)':  '17.4 as KMnO4'},
        {'temp': 10,   'Alkalinity':    '0 as HCO3', 'O(0)':  '1 ', 'C(-4)':    '0 as CH4', 'pH': 4.4, 'Na':   '19', 'K':   1.8, 'Ca':   2, 'Mg':   3.0, 'Fe': 0.12, 'Mn': 0.03, 'Amm':  0.05, 'Si': '10.6 as SiO2', 'Cl':   31, 'S(6)': '18 as SO4', 'N(5)': '12.0 as NO3', 'F':  0.07000, 'Mn(7)':   '3.0 as KMnO4'},
        {'temp': 10,   'Alkalinity':    '7 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 5.5, 'Na':   '20', 'K':   2.1, 'Ca':   3, 'Mg':   3.9, 'Fe': 2.33, 'Mn': 0.06, 'Amm':  0.08, 'Si': '15.4 as SiO2', 'Cl':   34, 'S(6)': '19 as SO4', 'N(5)':  '0.1 as NO3', 'F':  0.12000, 'Mn(7)':   '5.3 as KMnO4'},
        {'temp': 10,   'Alkalinity':   '13 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 5.7, 'Na':   '21', 'K':   2.8, 'Ca':   5, 'Mg':   5.4, 'Fe': 3.43, 'Mn': 0.13, 'Amm':  0.19, 'Si': '18.6 as SiO2', 'Cl':   41, 'S(6)': '20 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00501, 'Mn(7)':   '7.3 as KMnO4'},
        {'temp': 10,   'Alkalinity':   '23 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 6.2, 'Na':   '24', 'K':   1.8, 'Ca':  13, 'Mg':   2.7, 'Fe': 0.70, 'Mn': 0.03, 'Amm':  0.23, 'Si': '15.1 as SiO2', 'Cl':   41, 'S(6)': '24 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00501, 'Mn(7)':  '13.9 as KMnO4'},
        {'temp': 10,   'Alkalinity':   '92 as HCO3', 'O(0)':  '0 ', 'C(-4)': '0.14 as CH4', 'pH': 7.4, 'Na':   '29', 'K':   1.9, 'Ca':  38, 'Mg':   2.8, 'Fe': 0.30, 'Mn': 0.09, 'Amm':  0.24, 'Si': '15.6 as SiO2', 'Cl':   44, 'S(6)': '36 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00501, 'Mn(7)':  '16.8 as KMnO4'},
        {'temp': 10,   'Alkalinity':  '121 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 7.6, 'Na':   '31', 'K':   2.0, 'Ca':  47, 'Mg':   3.3, 'Fe': 0.40, 'Mn': 0.13, 'Amm':  0.34, 'Si': '13.3 as SiO2', 'Cl':   50, 'S(6)': '35 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00501, 'Mn(7)':  '11.9 as KMnO4'},
        {'temp': 10.1, 'Alkalinity':  '146 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 7.7, 'Na':   '29', 'K':   2.0, 'Ca':  51, 'Mg':   3.4, 'Fe': 0.51, 'Mn': 0.13, 'Amm':  0.44, 'Si': '20.0 as SiO2', 'Cl':   45, 'S(6)': '22 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.01000, 'Mn(7)':   '0.0 as KMnO4'},
        {'temp': 10.2, 'Alkalinity':  '151 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 7.7, 'Na':   '15', 'K':   1.4, 'Ca':  48, 'Mg':   3.1, 'Fe': 3.40, 'Mn': 0.18, 'Amm':  0.68, 'Si': '16.0 as SiO2', 'Cl':   33, 'S(6)':  '0 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.01000, 'Mn(7)':   '0.0 as KMnO4'},
        {'temp': 10.5, 'Alkalinity':  '156 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '5 as CH4', 'pH': 7.6, 'Na':   '15', 'K':   1.2, 'Ca':  48, 'Mg':   2.5, 'Fe': 1.22, 'Mn': 0.13, 'Amm':  0.45, 'Si': '17.6 as SiO2', 'Cl':   29, 'S(6)':  '0 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.03000, 'Mn(7)':  '16.6 as KMnO4'},
        {'temp': 10.7, 'Alkalinity':  '169 as HCO3', 'O(0)':  '0 ', 'C(-4)':  '9.1 as CH4', 'pH': 8.1, 'Na':   '24', 'K':   9.5, 'Ca':  13, 'Mg':  21.0, 'Fe': 0.69, 'Mn': 0.04, 'Amm':  0.83, 'Si': '22.9 as SiO2', 'Cl':   33, 'S(6)':  '0 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.10000, 'Mn(7)':  '11.5 as KMnO4'},
        {'temp': 10.8, 'Alkalinity':  '398 as HCO3', 'O(0)':  '0 ', 'C(-4)': '18.2 as CH4', 'pH': 8.5, 'Na':  '155', 'K':  11.5, 'Ca':   3, 'Mg':   3.1, 'Fe': 0.62, 'Mn': 0.04, 'Amm':  6.19, 'Si': '17.4 as SiO2', 'Cl':   35, 'S(6)':  '0 as SO4', 'N(5)':  '0.0 as NO3', 'F':  1.50000, 'Mn(7)':  '84.9 as KMnO4'},
        {'temp': 10.7, 'Alkalinity':  '525 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 8.2, 'Na':  '190', 'K':  16.8, 'Ca':   9, 'Mg':   9.5, 'Fe': 0.17, 'Mn': 0.12, 'Amm':  8.26, 'Si': '45.0 as SiO2', 'Cl':   60, 'S(6)':  '0 as SO4', 'N(5)':  '0.0 as NO3', 'F':  1.10000, 'Mn(7)': '132.0 as KMnO4'},
        {'temp': 10.8, 'Alkalinity':  '801 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 7.7, 'Na':  '445', 'K':  58.0, 'Ca':  27, 'Mg':  19.9, 'Fe': 1.60, 'Mn': 0.12, 'Amm': 20.00, 'Si': '83.2 as SiO2', 'Cl':  390, 'S(6)':  '9 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00000, 'Mn(7)': '470.0 as KMnO4'},
        {'temp': 10.8, 'Alkalinity':  '903 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 8.5, 'Na':  '910', 'K': 134.0, 'Ca':  66, 'Mg':  71.0, 'Fe': 5.40, 'Mn': 0.30, 'Amm': 49.00, 'Si': '71.6 as SiO2', 'Cl': 1320, 'S(6)': '15 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00000, 'Mn(7)': '130.0 as KMnO4'},
        {'temp': 11.0, 'Alkalinity': '2635 as HCO3', 'O(0)':  '0 ', 'C(-4)':    '0 as CH4', 'pH': 7.8, 'Na': '3380', 'K':  95.0, 'Ca': 160, 'Mg': 390.0, 'Fe': 2.30, 'Mn': 0.03, 'Amm': 49.00, 'Si': '71.6 as SiO2', 'Cl': 5285, 'S(6)':  '4 as SO4', 'N(5)':  '0.0 as NO3', 'F':  0.00000, 'Mn(7)': '160.0 as KMnO4'},
    ]

    solutions = [None] * len(solution_dictionaries)
    for _i, sol_dict in enumerate(solution_dictionaries):
        sol_dict['units'] = 'mg/L'
        solutions[_i] = pp.add_solution(sol_dict)
    return pp, solutions


def test_data_is_valid(mineral_data):
    ''' test whether the data in the excel that contains all
        data is valid. The data is used to test all the phreeqpython
        related functionality '''
    assert mineral_data.hgc.is_valid


def test_add_not_consolidated_solution(mineral_data):
    ''' test the calculation of the saturation index (SI) of
        calcite is performed correctly '''
    df = mineral_data
    with pytest.raises(ValueError) as excinfo:
        df.hgc.get_phreeqpython_solutions()
        assert "The required column ph is missing in the dataframe.  " in excinfo


def test_phreeqpython_installed():
    '''test phreeqpython can be imported and used. test
       this by adding 1 simple solution to the PhreeqPython instance'''
    pp = PhreeqPython()
    pp.add_solution_simple({'CaCl2': 1.0, 'NaHCO3': 2.0})


def test_add_oxygen(consolidated_data, phreeqpython_solutions_excel):
    ''' Test oxygen is added and returned correctly to and from
    phreeqpython with or without. Test it in pure phreeqpython (because there
    it is not working as expected)
    - test that no O(0) added yields no O in solution
    - test that adding `O(0)` `as O2` yields wrong result
    - test that adding `O(0)` without `as O2` yields correct result
    - test that adding in the SamplesFrame O2 concentration yields correct results '''
    pp = PhreeqPython()
    test_solution_dict = {'units': 'mg/L',
                          'Cl': 8.0,
                          'Na': '2.0 charge',
                          'Alkalinity': '2.0 as HCO3',
                          'Ca': 1.0,
                          'pH': 7,
                          'temp': 11
                          }
    sol = pp.add_solution(test_solution_dict)
    assert sol.total_element('O') == 0.0

    test_solution_dict['O(0)'] = '3 as O2'
    sol1 = pp.add_solution(test_solution_dict)
    assert sol1.species['O2'] * 2 * mw('O') * 1000. != pytest.approx(3., 1.e-4)

    test_solution_dict['O(0)'] = 3
    sol1 = pp.add_solution(test_solution_dict)
    assert sol1.species['O2'] * 2 * mw('O') * 1000. == pytest.approx(3., 1.e-4)

    test_data = pd.DataFrame({'Cl': 8, 'Na': 2, 'alkalinity': 2,
                              'Ca': 1, 'O2': 3, 'ph': 7,
                              'temp': 11}, index=[0])
    test_data.hgc.make_valid()
    test_data.hgc.consolidate(
        use_so4=None, use_ph=None, use_ec=None, use_temp=None)
    sols = test_data.hgc.get_phreeqpython_solutions(inplace=False)
    assert sols[0].species['O2'] * 2 * \
        mw('O') * 1000. == pytest.approx(3., 1.e-4)


def test_warning_with_hco3_column(caplog):
    """ test whether a warning is logged when using hco3 as column (and not
    alkalinity). and whether hco3 and bicarbonate columns in SamplesFrame are
    ignored in phreeq calculations"""
    test_data_no_alk = pd.DataFrame({'Cl': 8, 'Na': 2,
                                     'Ca': 1, 'O2': 3, 'ph': 7,
                                     'temp': 11}, index=[0])
    test_data_alk = pd.DataFrame({'Cl': 8, 'Na': 2, 'alkalinity': 2,
                                  'Ca': 1, 'O2': 3, 'ph': 7,
                                  'temp': 11}, index=[0])
    test_data_hco3 = pd.DataFrame({'Cl': 8, 'Na': 2, 'hco3': 2,
                                   'Ca': 1, 'O2': 3, 'ph': 7,
                                   'temp': 11}, index=[0])
    test_data_bicarb = pd.DataFrame({'Cl': 8, 'Na': 2, 'bicarbonate': 2,
                                     'Ca': 1, 'O2': 3, 'ph': 7,
                                     'temp': 11}, index=[0])
    test_data_both = pd.DataFrame({'Cl': 8, 'Na': 2, 'bicarbonate': 2,
                                   'alkalinity': 2,
                                   'Ca': 1, 'O2': 3, 'ph': 7,
                                   'temp': 11}, index=[0])
    test_data_no_alk.hgc.make_valid()
    test_data_alk.hgc.make_valid()
    test_data_hco3.hgc.make_valid()
    test_data_bicarb.hgc.make_valid()
    test_data_both.hgc.make_valid()

    test_data_no_alk.hgc.consolidate(
        use_so4=None, use_ph=None, use_ec=None, use_temp=None)
    test_data_alk.hgc.consolidate(
        use_so4=None, use_ph=None, use_ec=None, use_temp=None)
    test_data_hco3.hgc.consolidate(
        use_so4=None, use_ph=None, use_ec=None, use_temp=None)
    test_data_bicarb.hgc.consolidate(
        use_so4=None, use_ph=None, use_ec=None, use_temp=None)
    test_data_both.hgc.consolidate(
        use_so4=None, use_ph=None, use_ec=None, use_temp=None)

    # capture all logs and compare the messages to the expected messages.
    caplog.clear()
    with caplog.at_level(logging.WARNING):
        sol_no_alk = test_data_no_alk.hgc.get_phreeqpython_solutions(inplace=False)
    assert len(caplog.records) == 0

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        sol_alk = test_data_alk.hgc.get_phreeqpython_solutions(inplace=False)
    assert len(caplog.records) == 0

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        sol_hco3 = test_data_hco3.hgc.get_phreeqpython_solutions(inplace=False)
    assert 'bicarbonate (or hco3) is found, but no alkalinity' in caplog.text

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        sol_bicarb = test_data_bicarb.hgc.get_phreeqpython_solutions(inplace=False)
    assert 'bicarbonate (or hco3) is found, but no alkalinity' in caplog.text

    caplog.clear()
    with caplog.at_level(logging.WARNING):
        sol_both = test_data_both.hgc.get_phreeqpython_solutions(inplace=False)
    assert 'bicarbonate (or hco3) and alkalinity' in caplog.text

    assert sol_hco3.values[0].species == sol_bicarb.values[0].species
    assert sol_hco3.values[0].species == sol_no_alk.values[0].species
    assert sol_hco3.values[0].species != sol_alk.values[0].species
    assert sol_both.values[0].species == sol_alk.values[0].species


def test_get_phreeqpython_solutions(consolidated_data, phreeqpython_solutions_excel):
    ''' Assert phreeqpython solutions are returned as series with correct
       compounds'''
    df = consolidated_data
    # pp = phreeqpython_solutions_excel[0]
    solutions_direct = phreeqpython_solutions_excel[1]

    solutions_hgc = df.hgc.get_phreeqpython_solutions(inplace=False)
    assert all([isinstance(s, Solution) for s in solutions_hgc])

    for _i, sol in enumerate(solutions_hgc):
        sol_pp = solutions_direct[_i]
        assert sol.species_molalities == sol_pp.species_molalities, f'species molalities are not equal for solution #{_i}'
        # assert sol.species_molalities == pytest.approx(sol_pp.species_molalities, abs=1.e-4, rel=1e-1), f'species molalities are not equal for solution #{_i}'


def test_solution_auto_equilibrate(consolidated_data):
    """ assert that solutions are equilibrated with Na or Cl depending
        on what their concentrations are """
    df = consolidated_data
    df = df.iloc[:3, :]
    # make sure that equilibrating with Na will fail because
    # it will needs a concentration below 0 to match the
    # charge in balance
    df.loc[0, 'Fe'] = 2*df.loc[0, 'Na']
    sol = df.hgc.get_phreeqpython_solutions(equilibrate_with='auto', inplace=False)
    Na_in_sol = [s.total_element('Na') * mw('Na') for s in sol]
    Cl_in_sol = [s.total_element('Cl') * mw('Cl') for s in sol]
    # first one equilibrates with Cl
    np.testing.assert_allclose(
        Na_in_sol[:1], df.loc[:0, 'Na'].values, rtol=1.e-1)
    with pytest.raises(AssertionError):
        np.testing.assert_allclose(Na_in_sol[1:],
                                   df.loc[1:, 'Na'].values,
                                   rtol=1.e-1)
    # others one equilibrates with Na
    np.testing.assert_allclose(
        Cl_in_sol[1:], df.loc[1:, 'Cl'].values, rtol=1.e-1)


def test_solution_equilibrate_with(consolidated_data):
    ''' Assert phreeqpython solutions are returned as series'''
    df = consolidated_data

    solutions_default = df.hgc.get_phreeqpython_solutions(inplace=False)
    solutions_Na = df.hgc.get_phreeqpython_solutions(equilibrate_with='Na', inplace=False)
    solutions_Cl = df.hgc.get_phreeqpython_solutions(equilibrate_with='Cl', inplace=False)
    solutions_none = df.hgc.get_phreeqpython_solutions(equilibrate_with=None, inplace=False)
    solutions_auto = df.hgc.get_phreeqpython_solutions(equilibrate_with='auto', inplace=False)

    # get the list of Na-concentrations in the phreeqpython-solutions (from mmol/L to
    # mg/L)
    Na_in_sol_default = [s.total_element('Na') * mw('Na')
                         for s in solutions_default]
    Na_in_sol_Na = [s.total_element('Na') * mw('Na') for s in solutions_Na]
    Na_in_sol_Cl = [s.total_element('Na') * mw('Na') for s in solutions_Cl]
    Na_in_sol_none = [s.total_element('Na') * mw('Na') for s in solutions_none]
    Na_in_sol_auto = [s.total_element('Na') * mw('Na') for s in solutions_auto]

    # get the list of Fe-concentrations in the phreeqpython-solutions (from mmol/L to
    # mg/L)
    Fe_in_sol_default = [s.total_element('Fe') * mw('Fe')
                         for s in solutions_default]
    Fe_in_sol_none = [s.total_element('Fe') * mw('Fe')
                      for s in solutions_none]
    Fe_in_sol_auto = [s.total_element('Fe') * mw('Fe')
                      for s in solutions_auto]

    # test that default equilibrate with is None by returning the same array
    np.testing.assert_array_equal(Na_in_sol_default, Na_in_sol_none)
    # assert that indeed Na-concentration is altered by comparing it
    # to the Na concentration when Cl is used for equilibration
    assert all(np.not_equal(Na_in_sol_Cl, Na_in_sol_Na))
    # assert that auto equals with Na for this case
    assert all(np.not_equal(Na_in_sol_none, Na_in_sol_Na))
    # assert Na-concentration is not altered by using equilibrate with
    # Cl. Na concentration should be the same as in the original dataframe
    # to some extent of accuracy (this is generally not closer than 10% in my experience).
    np.testing.assert_allclose(Na_in_sol_Cl, df.Na.values, rtol=1.e-1)
    np.testing.assert_allclose(Na_in_sol_none, df.Na.values, rtol=1.e-1)
    # same assertion but now for Fe
    np.testing.assert_allclose(Fe_in_sol_default, df.Fe.values, rtol=1.e-1)
    np.testing.assert_allclose(Fe_in_sol_none, df.Fe.values, rtol=1.e-1)
    np.testing.assert_allclose(Fe_in_sol_auto, df.Fe.values, rtol=1.e-1)


# def test_calculate_ec_with_measured_values(mineral_data):
#     ''' test whether ec is calculated 10% accurately by hgc compared
#         to the measured values given in the original excel-hgc example
#         file '''
#     df = mineral_data
#     df.hgc.consolidate(use_ph='lab')
#     ec = df.hgc.get_specific_conductance()

#     # we cannot use data of the HGC excel sheet to check
#     # whether the correct values have been calculated,
#     # because Pieter has different ways of calculating it.
#     np.testing.assert_allclose(ec, df.ec.values, rtol=1.e-1)

#     # test it by calculating it seperately by phreeqpython manually.


def test_get_specific_conductance(consolidated_data, phreeqpython_solutions_excel):
    ''' Assert get_specific_conductance wrapper returns correct saturation indices of all test
        solutions in the fixtures '''
    df = consolidated_data
    solutions_direct = phreeqpython_solutions_excel[1]
    sc_hgc = df.hgc.get_specific_conductance(inplace=False)
    sc_pp = pd.Series([sol.sc for sol in solutions_direct], index=sc_hgc.index, name=sc_hgc.name)

    pd.testing.assert_series_equal(sc_hgc, sc_pp)


def test_get_saturation_index_unknown_mineral(consolidated_data, phreeqpython_solutions_excel, caplog):
    """ Assert if saturation of unkown mineral is -Inf (i.e. <-900)"""
    df = consolidated_data
    si_unknown = df.hgc.get_saturation_index('Unknown', inplace=False)
    assert si_unknown.min() < 900

def test_get_saturation_index_mineral_with_brackets(consolidated_data, phreeqpython_solutions_excel, caplog):
    """ PHREEQC allows very strange names for minerals or phases. E.g. [NH4] CO2(g)"""
    df = consolidated_data
    si_nh4 = df.hgc.get_saturation_index('[NH4]', inplace=False)
    assert si_nh4.min() < 900
    si_co2 = df.hgc.get_saturation_index('CO2(g)', inplace=False)
    # waters without CO2 or HCO3 should have -Inf (=-999) as saturation index
    assert si_co2.min() < 900
    # waters with CO2/HCO3 should have a saturation index above -10 but below 0
    assert all(si_co2[3:] >-10)
    assert all(si_co2[3:] <0)
    df.hgc.get_saturation_index('CO2(g)', inplace=True)
    assert all(df['si_co2(g)'] == si_co2)

def test_get_partial_pressure(consolidated_data):
    """ Same as test_get_saturation_index as it is only an alias"""
    df = consolidated_data
    pp_co2 = df.hgc.get_partial_pressure('CO2(g)', inplace=False)
    # waters without CO2 or HCO3 should have -Inf (=-999) as saturation index
    assert pp_co2.min() < 900
    # waters with CO2/HCO3 should have a saturation index above -10 but below 0
    assert all(pp_co2[3:] >-10)
    assert all(pp_co2[3:] <0)
    df.hgc.get_partial_pressure('CO2(g)', inplace=True)
    assert all(df['pp_co2(g)'] == pp_co2)

def test_get_saturation_index_calcite(consolidated_data, phreeqpython_solutions_excel, caplog):
    ''' Assert get_si wrapper returns correct saturation indices of all test
        solutions in the fixtures '''
    df = consolidated_data
    # pp = phreeqpython_solutions_excel[0]
    solutions_direct = phreeqpython_solutions_excel[1]

    si_calcite_hgc = df.hgc.get_saturation_index('Calcite', inplace=False)
    si_calcite_pp = pd.Series([sol.si('Calcite') for sol in solutions_direct],
                              index=si_calcite_hgc.index,
                              name=si_calcite_hgc.name)

    pd.testing.assert_series_equal(si_calcite_hgc, si_calcite_pp)

    df.drop(columns=['doc','DOC'], errors='ignore', inplace=True)

    caplog.clear()
    with caplog.at_level(logging.INFO):
        df.hgc.get_saturation_index('Calcite', inplace=True)
    assert len(caplog.records) == 1
    column_name = 'si_calcite'
    assert column_name in caplog.text.lower()
    assert f'{column_name}' in set(df.columns)


import logging
from hgc.samples_frame import SamplesFrame
from hgc.constants.constants import mw
import pandas as pd
import numpy as np
from unittest import TestCase, mock
from datetime import datetime
import pytest
from . import test_directory


# define the fixtures
@pytest.fixture(name='test_data_bas_vdg')
def fixture_test_bas_vdg():
    """ test data as been used by Bas vdG from testing routine 060602020"""
    test_data = {
        'ph_lab': [7.5, 6.1, 7.6], 'ph_field': [4.4, 6.1, 7.7],
        'ec_lab': [304, 401, 340], 'ec_field': [290, 'error', 334.6],
        'temp': [10, 10, 10],
        'alkalinity':  [110, 7, 121],
        'O2':  [11, 0, 0],
        'Na': [2,40,310],
        'K':[0.4, 2.1, 2.0],
        'Ca':[40,3,47],
        'Fe': [0.10, 2.33, 0.4],
        'Mn': [0.02, 0.06, 0.13],
        'NH4': [1.29, 0.08, 0.34],
        'SiO2': [0.2, 15.4, 13.3],
        'SO4': [7,19,35],
        'NO3': [3.4,0.1,0],
        'Cl': [10,50,310]
    }
    df = pd.DataFrame(test_data)
    df.hgc.make_valid()
    return pd.DataFrame(df)

@pytest.fixture(name='test_data_bas_vdg_consolidated')
def fixture_test_bas_vdg_consolidated(test_data_bas_vdg):
    df = test_data_bas_vdg.copy()
    df.hgc.consolidate(use_ph='lab', use_ec='lab', use_temp=None, use_so4=None, use_o2=None)
    return df

def test_valid_samples_frame():
    df = pd.read_csv(test_directory / 'data' / 'dataset_basic.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)
    assert df.hgc.is_valid == True

def test_invalid_changed_samples_frame():
    df = pd.read_csv(test_directory / 'data' / 'dataset_basic.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)
    assert df.hgc.is_valid == True
    df.loc[1, 'F'] = -1
    # this test does not fail because the self._obj in df.hgc is the orginal
    # pandas obj and is not changed in the hgc namespace
    assert df.hgc._obj.loc[1, 'F'] == -1
    assert df.hgc.is_valid == False

def test_valid_samples_frame_excel():
    df = pd.read_excel(test_directory / 'data' / 'dataset_basic.xlsx', skiprows=[1])
    assert df.hgc.is_valid == True

def test_invalid_samples_frame():
    df = pd.read_csv(test_directory / 'data' / 'dataset_invalid_columns.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)
    assert df.hgc.is_valid == False


def test_make_valid():
    df = pd.read_csv(test_directory / 'data' / 'dataset_invalid_columns.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)
    df.hgc.make_valid()

    assert df.hgc.is_valid == True

def test_replace_detection_limit():
    df_original = pd.read_csv(test_directory / 'data' / 'dataset_invalid_columns.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)
    df_half = df_original.copy(deep=True)
    df_on = df_original.copy(deep=True)
    df_zero = df_original.copy(deep=True)
    df_invalid = df_original.copy(deep=True)

    df_half.hgc._replace_detection_lim(rule='half')
    df_on.hgc._replace_detection_lim(rule='on')
    df_zero.hgc._replace_detection_lim(rule='zero')

    with pytest.raises(ValueError) as exc_info:
        df_invalid.hgc._replace_detection_lim(rule='invalid_option')
    assert 'Invalid rule' in str(exc_info)

    expected_changed_column = 'ec_field'
    expected_unchanged_columns = list(set(df_half.columns) - {expected_changed_column})

    df_is_not_nan_and_not_changed_half = df_half.isna() | (df_half == df_original)
    df_is_not_nan_and_not_changed_on = df_on.isna() | (df_on == df_original)
    df_is_not_nan_and_not_changed_zero = df_zero.isna() | (df_zero == df_original)

    assert all(df_is_not_nan_and_not_changed_half[expected_unchanged_columns])
    assert all(df_is_not_nan_and_not_changed_on[expected_unchanged_columns])
    assert all(df_is_not_nan_and_not_changed_zero[expected_unchanged_columns])

    assert not all(df_half[expected_changed_column] == df_original[expected_changed_column])
    assert not all(df_on[expected_changed_column] == df_original[expected_changed_column])
    assert not all(df_zero[expected_changed_column] == df_original[expected_changed_column])

    assert all(df_half.ec_field.astype(float) == [200, 350, 215, 1.5, 1.5, 525])
    assert all(df_on.ec_field.astype(float) == [200, 350, 215, 3, 3, 525])
    assert all(df_zero.ec_field.astype(float) == [200, 350, 215, 0, 0, 525])


def test_get_ratios_invalid_frame():
    df = pd.DataFrame()
    with pytest.raises(ValueError):
        df.hgc.get_ratios()


def test_elemental_ratios():
    """ test all the ratios are calculated correctly"""

# Cl/Br (molar ratio)
#          * Cl/Na (molar ratio)
#          * Cl/Mg (molar ratio)
#          * Ca/Sr (molar ratio)
#          * Fe/Mn (molar ratio)
#          * HCO3/Ca
    ratios = dict(cl_to_br = ['Cl', 'Br'],
                  cl_to_na = ['Cl', 'Na'],
                  ca_to_mg = ['Ca', 'Mg'],
                  ca_to_sr = ['Ca', 'Sr'],
                  fe_to_mn = ['Fe', 'Mn'],
                  )
    for ratio, (numerator, denominator) in ratios.items():
        dict_df = {
            'ph': np.array(6 * [7]),
            numerator:np.array([1, 2, 3, 1, 2, 3]) * mw(numerator),
            denominator:np.array([1, 1, 1, 2, 2, 2]) * mw(denominator)}
        # dict_df = {k: mw(k)*conc for k, conc in dict_df.items()}
        df = pd.DataFrame.from_dict(dict_df)
        df.hgc.get_ratios()
        np.testing.assert_array_almost_equal(df[ratio].values, np.array([1, 2, 3, 0.5, 1, 1.5]))

def test_alkalinity_ratios():

    dict_df = {
        'alkalinity':np.array([1, 2, 3, 1, 2, 3]) * mw('HCO3'),
        'Ca':np.array([1, 1, 1, 2, 2, 2]) * mw('Ca'),
        'Mg':np.array([1, 1, 1, 2, 2, 2]) * mw('Mg'),
        'ph':np.array(6 * [7]),
    }
    df = pd.DataFrame.from_dict(dict_df)
    df.hgc.get_ratios()

    np.testing.assert_array_almost_equal(df['hco3_to_ca'].values, np.array([1, 2, 3, 0.5, 1, 1.5]))
    np.testing.assert_array_almost_equal(df['hco3_to_ca_and_mg'].values, np.array([0.5, 1, 1.5, 0.25, 0.5, .75]))
    np.testing.assert_array_almost_equal(df['hco3_to_sum_anions'].values, np.array(6 * [1]))

    dict_df['Cl'] = np.array([4, 4, 4, 8, 8, 8]) * mw('Cl')
    df = pd.DataFrame.from_dict(dict_df)
    df.hgc.get_ratios()
    np.testing.assert_array_almost_equal(df['hco3_to_ca'].values, np.array([1, 2, 3, 0.5, 1, 1.5]))
    np.testing.assert_array_almost_equal(df['hco3_to_ca_and_mg'].values, np.array([0.5, 1, 1.5, 0.25, 0.5, .75]))
    np.testing.assert_array_almost_equal(df['hco3_to_sum_anions'].values, np.array([1./5, 2./6, 3./7, 1./9, 2./10, 3./11]))


def test_get_ratios_larger_df(caplog):
    df = pd.read_csv(test_directory / 'data' / 'dataset_basic.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)

    df.hgc.consolidate(use_ph='lab')
    df_ratios = df.hgc.get_ratios(inplace=False)
    assert isinstance(df_ratios, pd.core.frame.DataFrame)

    caplog.clear()
    with caplog.at_level(logging.INFO):
        df.hgc.get_ratios()
    assert len(caplog.records) > 0

def test_consolidate():
    df = pd.read_csv(test_directory / 'data' / 'dataset_basic.csv', skiprows=[1], parse_dates=['date'], dayfirst=True)
    df.hgc.consolidate(use_so4=None, use_o2=None, use_ph='lab')

def test_consolidate_w_not_all_cols():
    ''' test that consolidate works when not all
        (default) columns are present '''
    testdata = {
        'ph_lab': [4.3, 6.3, 5.4], 'ph_field': [4.4, 6.1, 5.7],
        'ec_lab': [304, 401, 340], 'ec_field': [290, 'error', 334.6],
    }
    df = pd.DataFrame.from_dict(testdata)

    df.hgc.make_valid()

    with pytest.raises(ValueError):
        df.hgc.consolidate(use_ph='field', use_ec='lab',)

    df.hgc.consolidate(use_ph='field', use_ec='lab', use_temp=None,
                       use_so4=None, use_o2=None)

def test_consolidate_alkalinity():
    ''' test that consolidate works when not all
        (default) columns are present '''
    testdata = {
        'alkalinity': [4.3, 6.3, 5.4], 'hco3': [4.4, 6.1, 5.7],
        'ph_lab': [4.3, 6.3, 5.4], 'ph_field': [4.4, 6.1, 5.7],
        'ec_lab': [304, 401, 340], 'ec_field': [290, 'error', 334.6],
    }
    df = pd.DataFrame.from_dict(testdata)
    df.hgc.make_valid()

    alk = df.alkalinity
    hco3 = df.hco3
    df.hgc.consolidate(use_ph='field', use_ec='lab',use_alkalinity='alkalinity',
    use_so4=None, use_o2=None, use_temp=None)
    pd.testing.assert_series_equal(df.alkalinity, alk)
    df.hgc.consolidate(use_alkalinity='hco3', use_so4=None, use_o2=None,
    use_temp=None)
    assert all(df.alkalinity.values == hco3.values)
    assert 'hco3' not in df.columns



def test_get_sum_anions_1():
    """ This testcase is based on row 11, sheet 4 of original Excel-based HGC """
    df = pd.DataFrame([[56., 16., 1.5, 0.027, 0.0, 0.0, 3.4, 0.04, 7., 4.5]], columns=('Br', 'Cl', 'doc', 'F', 'alkalinity', 'NO2', 'NO3', 'PO4', 'SO4', 'ph'))
    df.hgc.make_valid()
    sum_anions = df.hgc.get_sum_anions(inplace=False)
    assert np.round(sum_anions[0], 2)  == 0.67

def test_get_sum_anions_2():
    """ This testcase is based on sheet 5, row 12 of original Excel-based HGC """
    testdata = {
        'Br': [0],
        'Cl': [19.0],
        'doc': [4.4],
        'F': [0.08],
        'alkalinity': [0.0],
        'NO2': [0.0],
        'NO3': [22.6],
        'PO4': [0.04],
        'SO4': [16.0],
        'ph': [4.3]
    }
    df = pd.DataFrame.from_dict(testdata)
    df.hgc.make_valid()
    sum_anions = df.hgc.get_sum_anions(inplace=False)
    assert np.round(sum_anions[0], 2)  == 1.28

def test_get_sum_anions_3(test_data_bas_vdg):
    """ Test based on Bas vd Grift bug report """
    df = test_data_bas_vdg
    df.hgc.consolidate(use_ph='lab', use_ec='lab', use_temp=None, use_so4=None, use_o2=None)

    sum_anions = df.hgc.get_sum_anions(inplace=False)
    np.testing.assert_almost_equal(sum_anions.values,
                                   np.array([2.285, 1.922, 11.456]),
                                   decimal=3)

    sum_cations = df.hgc.get_sum_cations(inplace=False)
    np.testing.assert_almost_equal(sum_cations.values,
                                   np.array([2.1690812, 2.0341514, 15.9185133]))

def test_get_sum_cations():
    df = pd.DataFrame([[4.5, 9.0, 0.4, 1.0, 1.1, 0.1, 0.02, 1.29, 99.0, 3.0, 0.3, 3.2, 0.6, 0.6, 10.4, 7.0, 15.0]], columns=('ph', 'Na', 'K', 'Ca', 'Mg', 'Fe', 'Mn', 'NH4', 'Al', 'Ba', 'Co', 'Cu', 'Li', 'Ni', 'Pb', 'Sr', 'Zn'))
    df.hgc.make_valid()
    sum_cations = df.hgc.get_sum_cations(inplace=False)
    assert np.round(sum_cations[0], 2)  == 0.66

def test_get_sum_cations_no_cations():
    df = pd.DataFrame(dict(Cl=[4, 4], Na=[0,1], ph=[10,10]))
    df.hgc.make_valid()
    sum_cations = df.hgc.get_sum_cations(inplace=False)
    assert np.round(sum_cations[0], 2)  == 0.0
    assert np.round(sum_cations[1], 2)  > 0

def test_get_sum_anions_no_anions():
    df = pd.DataFrame(dict(Cl=[0, 4], Na=[1,1], ph=[10,10]))
    df.hgc.make_valid()
    sum_anions = df.hgc.get_sum_anions(inplace=False)
    assert np.round(sum_anions[0], 2)  == 0.0
    assert np.round(sum_anions[1], 2)  > 0

def test_get_sum_an_and_cations_no_ph_raises_error():
    df = pd.DataFrame(dict(Cl=[0, 4], Na=[1,1]))
    df.hgc.make_valid()
    with pytest.raises(ValueError) as exc_info:
        df.hgc.get_sum_anions(inplace=False)
    assert 'missing column ph' in str(exc_info).lower()

    with pytest.raises(ValueError) as exc_info:
        df.hgc.get_sum_cations(inplace=False)
    assert 'missing column ph' in str(exc_info).lower()

def test_get_sum_cations_invalid_ph():
    df = pd.DataFrame(dict(Cl=[4, 4], Na=[1,1], ph=[0, np.nan]))
    df.hgc.make_valid()
    with pytest.raises(ValueError) as exc_info:
        df.hgc.get_sum_anions(inplace=False)
    assert 'invalid value(s) in ph column' in str(exc_info).lower()

def test_get_ion_balance(test_data_bas_vdg_consolidated):
    expected_value = pytest.approx([-2.61, 2.82, 16.30], abs=0.01)
    assert test_data_bas_vdg_consolidated.hgc.get_ion_balance(inplace=False).to_numpy() == expected_value
    test_data_bas_vdg_consolidated.hgc.get_ion_balance(inplace=True)
    assert test_data_bas_vdg_consolidated.ion_balance.to_numpy() == expected_value


def test_get_stuyfzand_water_type():
    """ Testcase matches row 12, sheet 6 of HGC Excel """
    testdata = {
        'Al': [2600],
        'Ba': [44.0],
        'Br': [0.0],
        'Ca': [2.0],
        'Cl': [19.0],
        'Co': [1.2],
        'Cu': [4.0],
        'doc': [4.4],
        'F': [0.08],
        'Fe': [0.29],
        'alkalinity': [0.0],
        'K': [1.1],
        'Li': [5.0],
        'Mg': [1.6],
        'Mn': [0.05],
        'Na': [15.0],
        'Ni': [7.0],
        'NH4': [0.05],
        'NO2': [0.0],
        'NO3': [22.6],
        'Pb': [2.7],
        'PO4': [0.04],
        'ph': [4.3],
        'SO4': [16.0],
        'Sr': [50],
        'Zn': [60.0]
    }
    df = pd.DataFrame.from_dict(testdata)
    df.hgc.make_valid()
    water_type = df.hgc.get_stuyfzand_water_type(inplace=False)
    assert water_type[0] == 'g*NaNO3o'
    df.hgc.get_stuyfzand_water_type()
    assert df.water_type[0] == 'g*NaNO3o'

def test_get_stuyfzand_water_type_2(test_data_bas_vdg_consolidated):
    """ test based on bas van de grift his test data """
    # abbrevation
    df = test_data_bas_vdg_consolidated.copy()
    df['Mg'] = 0 # column required for BEX
    assert df.hgc.get_stuyfzand_water_type(inplace=False).to_list() == ['g1CaHCO3o', 'F*NaClo', 'B1NaCl']

    testdata = {
        'ph_lab': [7.5, 6.1, 7.6], 'ph_field': [4.4, 6.1, 7.7],
        'ec_lab': [304, 401, 340], 'ec_field': [290, 'error', 334.6],
        'temp': [10, 10, 10],
        'alkalinity':  [110, 7, 121],
        #  'HCO3':  [110, 7, 121],
        'O2':  [11, 0, 0],
        'Na': [2,40,310],
        'K':[0.4, 2.1, 2.0],
        'Ca':[40,3,47],
        'Fe': [0.10, 2.33, 0.4],
        'Mn': [0.02, 0.06, 0.13],
        'NH4': [1.29, 0.08, 0.34],
        #  'Amm': [1.29, 0.08, 0.34],
        'SiO2': [0.2, 15.4, 13.3],
        'SO4': [7,19,35],
        'NO3': [3.4,0.1,0],
        'Cl': [10,50,310]
    }

    df = pd.DataFrame.from_dict(testdata)
    df.hgc.make_valid()
    df.hgc.consolidate(use_ph='lab', use_ec='lab', use_temp=None, use_so4=None, use_o2=None)
    df['Mg'] = 0
    assert df.hgc.get_stuyfzand_water_type(inplace=False).to_list() == ['g1CaHCO3o', 'F*NaClo', 'B1NaCl']



def test_get_bex():
    """ Sheet 5 - col EC in HGC Excel """
    df = pd.DataFrame([[15., 1.1, 1.6, 19.]], columns=('Na', 'K', 'Mg', 'Cl'))
    df.hgc.make_valid()
    bex = df.hgc.get_bex(inplace=False)
    assert np.round(bex[0], 2)  == 0.24

def test_get_bex_while_cl_has_nan():
    """ no outcome if any of the columns has no valid value    """
    df = pd.DataFrame([[15., 1.1, 1.6, 19],
                       [15., 1.1, 1.6, np.nan]], columns=('Na', 'K', 'Mg', 'Cl'))
    df.hgc.make_valid()
    with pytest.raises(ValueError) as exc_info:
        bex = df.hgc.get_bex(inplace=False)
    assert "column(s) ['cl" in str(exc_info).lower()

def test_get_bex_without_cl_column():
    """ no outcome if any of the columns has no valid value    """
    df = pd.DataFrame([[15., 1.1, 1.6 ],
                       [15., 1.1, 1.6]], columns=('Na', 'K', 'Mg'))
    df.hgc.make_valid()
    with pytest.raises(ValueError) as exc_info:
        bex = df.hgc.get_bex(inplace=False)
    assert "column(s) ['cl" in str(exc_info).lower()

def test_inplace(test_data_bas_vdg):
    """ Test to see if the inplace argument behaves as expected: returning
    a series in inplace=False and appending the column if inplace=True"""
    test_data_bas_vdg.hgc.consolidate(use_so4=None, use_o2=None, use_ph='lab')
    test_data_bas_vdg['Mg']=0
    def assert_column_added_inplace(column, is_added, method_name, method_kwargs):
        """ assert whether a column is added to the dataframe or not when
        a method with its arguments method_kwargs are called """
        df = test_data_bas_vdg.copy(deep=True)
        n_columns = len(df.columns)
        assert column not in df.columns
        method = getattr(df.hgc, method_name)
        df_out = method(**method_kwargs)
        if is_added:
            assert column in df.columns
            assert n_columns != len(df.columns)
            assert df_out is None
        else:
            assert column not in df.columns
            # assert n_columns == len(df.columns)  # this test does not work for bex because sum_anions is added in that case too. the test above should suffice anyway, so drop this test
            assert df_out is not None


    assert_column_added_inplace('bex', is_added=True, method_name='get_bex',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('bex', is_added=True, method_name='get_bex',
                                method_kwargs=dict())
    assert_column_added_inplace('bex', is_added=False, method_name='get_bex',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('cl_to_na', is_added=True, method_name='get_ratios',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('cl_to_na', is_added=True, method_name='get_ratios',
                                method_kwargs=dict())
    assert_column_added_inplace('cl_to_na', is_added=False, method_name='get_ratios',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('water_type', is_added=True, method_name='get_stuyfzand_water_type',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('water_type', is_added=True, method_name='get_stuyfzand_water_type',
                                method_kwargs=dict())
    assert_column_added_inplace('water_type', is_added=False, method_name='get_stuyfzand_water_type',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('dominant_anion', is_added=True, method_name='get_dominant_anions',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('dominant_anion', is_added=True, method_name='get_dominant_anions',
                                method_kwargs=dict())
    assert_column_added_inplace('dominant_anion', is_added=False, method_name='get_dominant_anions',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('dominant_cation', is_added=True, method_name='get_dominant_cations',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('dominant_cation', is_added=True, method_name='get_dominant_cations',
                                method_kwargs=dict())
    assert_column_added_inplace('dominant_cation', is_added=False, method_name='get_dominant_cations',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('sum_anions', is_added=True, method_name='get_sum_anions',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('sum_anions', is_added=True, method_name='get_sum_anions',
                                method_kwargs=dict())
    assert_column_added_inplace('sum_anions', is_added=False, method_name='get_sum_anions',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('sum_cations', is_added=True, method_name='get_sum_cations',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('sum_cations', is_added=True, method_name='get_sum_cations',
                                method_kwargs=dict())
    assert_column_added_inplace('sum_cations', is_added=False, method_name='get_sum_cations',
                                method_kwargs=dict(inplace=False))


    assert_column_added_inplace('pp_solutions', is_added=True, method_name='get_phreeqpython_solutions',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('pp_solutions', is_added=True, method_name='get_phreeqpython_solutions',
                                method_kwargs=dict())
    assert_column_added_inplace('pp_solutions', is_added=False, method_name='get_phreeqpython_solutions',
                                method_kwargs=dict(inplace=False))

    assert_column_added_inplace('si_calcite', is_added=True, method_name='get_saturation_index',
                                method_kwargs=dict(inplace=True, mineral_or_gas='calcite'))
    assert_column_added_inplace('si_calcite', is_added=True, method_name='get_saturation_index',
                                method_kwargs=dict(mineral_or_gas='calcite'))
    assert_column_added_inplace('si_calcite', is_added=False, method_name='get_saturation_index',
                                method_kwargs=dict(inplace=False, mineral_or_gas='calcite'))

    assert_column_added_inplace('sc', is_added=True, method_name='get_specific_conductance',
                                method_kwargs=dict(inplace=True))
    assert_column_added_inplace('sc', is_added=True, method_name='get_specific_conductance',
                                method_kwargs=dict())
    assert_column_added_inplace('sc', is_added=False, method_name='get_specific_conductance',
                                method_kwargs=dict(inplace=False))

def test_add_temp_ph_later():

    test_data = {
        'ph_lab': [7.5, 6.1, 7.6], 'ph_field': [4.4, 6.1, 7.7],
        'ec_lab': [304, 401, 340], 'ec_field': [290, 'error', 334.6],
        'alkalinity':  [110, 7, 121],
        'O2':  [11, 0, 0],
        'Na': [2,40,310],
        'K':[0.4, 2.1, 2.0],
        'Ca':[40,3,47],
        'Fe': [0.10, 2.33, 0.4],
        'Mn': [0.02, 0.06, 0.13],
        'NH4': [1.29, 0.08, 0.34],
        'SiO2': [0.2, 15.4, 13.3],
        'SO4': [7,19,35],
        'NO3': [3.4,0.1,0],
        'Cl': [10,50,310]
    }
    df = pd.DataFrame(test_data)
    df.hgc.make_valid()

    with pytest.raises(ValueError) as exc_info:
        df.hgc.get_saturation_index('Calcite')
    assert 'missing column ph' in str(exc_info).lower()
    df['ph']=7
    with pytest.raises(ValueError) as exc_info:
        df.hgc.get_saturation_index('Calcite')
    assert 'required column temp' in str(exc_info)
    df['temp']=10
    # assert no error is raised
    df.hgc.get_saturation_index('Calcite')







# -*- coding: utf-8 -*-
"""
   Test file that loads a big excel file and see if HGC's crashes or not
"""

import pandas as pd
from pathlib import Path
import hgc
from hgc.constants import constants

def test_hupsel():
    """ A test of a whole lot of samples that have been analysed manually with
    PHREEQC. The outcomes are not compared to the expected values. It only checks
     if no errors arise. """

    fpath = Path(__file__).parent / 'data'
    fname = 'analyse_bas.xls'

    df = pd.read_excel(
        io=fpath / fname,
        sheet_name='labab',
        skiprows= [0,1,3],
        usecols = ('A:L,N,O,Q,S,T,W:Y,AA:BJ'),
    )

    df = df.set_index('labcode')

    ####Mn van ug/L naar mg/L
    df.Mn = df.Mn/1000
    df.P = df.P/1000

    ####kolom namen aanpassen naar phreeqc: HCO3 naar alkalinity
    df =  df.rename(columns={"HCO3":"alkalinity",'pH':'ph','EC':'ec','P':'PO4'})

    ####NA invullen
    df['temp']=10
    values = {"ph": 7, "alkalinity": 10}
    df = df.fillna(value=values)


    df.hgc.make_valid()
    hgc_cols = df.hgc.hgc_cols
    print(hgc_cols)

    print([*constants.atoms])
    print([*constants.ions])
    print([*constants.properties])
    constants.atoms['Zn']

    df.hgc.get_sum_anions()
    df.hgc.get_sum_cations()

    df.hgc.get_dominant_anions(inplace=True)
    df.hgc.get_dominant_cations(inplace=True)
    df.hgc.get_ratios()
    df.hgc.get_stuyfzand_water_type()


    df.hgc.get_saturation_index('CO2(g)')
    df.hgc.get_saturation_index('Calcite',inplace=True)
    df.hgc.get_saturation_index('Calcite',inplace=False)
    df.hgc.get_specific_conductance()
    df.hgc.get_phreeqpython_solutions()
    df.pp_solutions.iloc[0]
    try:
        df.to_excel(Path(__file__).parent / 'data' / 'test.xlsx')
    except TypeError as e:
        if 'can only concatenate' in e.args[0]:
            # due to upstream syntax error in phreeqpython.
            pass
        else:
            raise e



# This is necessary so you can use `import hgc` and then
# all dataframes created afterwards are extended with the
# hgc namespace
from hgc.samples_frame import SamplesFrame

name = "hgc"

# allow access to mw and units via a shortcut hgc.mw
from hgc.constants.constants import mw, units

import logging
logging.basicConfig(format='%(levelname)s: %(message)s')
logging.getLogger().setLevel(logging.INFO)


"""
The `SamplesFrame` class is an extended Pandas DataFrame, offering additional methods
for validation of hydrochemical data, calculation of relevant ratios and classifications.

.. # an alias used in different doc strings in this file
.. |HCO3| replace:: HCO\ :sub:`3`\ :sup:`-`

"""
import logging

import numpy as np
import pandas as pd
from phreeqpython import PhreeqPython

from hgc.constants import constants
from hgc.constants.constants import mw

def requires_ph(func):
    """ Decorator function for methods in the SamplesFrame class that require a
    column `ph` with valid values (non-zero and non-NaN). """
    def wrapper(self, *args, **kwargs):
        if 'ph' not in self._obj.columns:#self._obj.columns:
            raise ValueError("Missing column ph. please use the consolidate method to use lab or field data on your dataframe or add a ph column manually with df['ph'] = some_ph")
        self._obj.loc[:, 'ph'] = self._obj['ph'].astype(float)
        isna_index = self._obj.ph.isna()
        iszero_index = self._obj.ph == 0
        if (any(isna_index) or any(iszero_index)):
            err_msg = "Invalid value(s) in ph column: "
            if any(isna_index):
                err_msg += f"NaN value(s) in row(s) {self._obj.index[isna_index].values}; "
            else:
                err_msg += f"no NaN; "

            if any(iszero_index):
                err_msg += f"Zero (0) value(s) in row(s) {self._obj.index[iszero_index].values}; "

            raise ValueError(err_msg)
        return func(self, *args, **kwargs)
    return wrapper

@pd.api.extensions.register_dataframe_accessor("hgc")
class SamplesFrame(object):
    """
    DataFrame with additional hydrochemistry-specific methods.
    All HGC methods and attributes defined in this class are available
    in the namespace 'hgc' of the Dataframe.

    Examples
    --------
    To use HGC methods, we always start from a Pandas DataFrame::

        import pandas as pd
        import hgc

        # We start off with an ordinary DataFrame
        df = pd.DataFrame({'Cl': [1,2,3], 'Mg': [11,12,13]})

        # Since we imported hgc, the HGC-methods become available
        # on the DataFrame. This allows for instance to use HGC's
        # validation function
        df.hgc.is_valid
        False
        df.hgc.make_valid()


    """
    __SUM_ANIONS_COLUMN =  'sum_anions'

    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self._pp = PhreeqPython() # bind 1 phreeqpython instance to the dataframe
        self._valid_atoms = constants.atoms
        self._valid_ions = constants.ions
        self._valid_properties = constants.properties

    @staticmethod
    def _clean_up_phreeqpython_solutions(solutions):
        """
        This is a convenience function that removes all
        the phreeqpython solution in `solutions` from
        memory.

        Parameters
        ----------
        solutions : list
            python list containing of phreeqpython solutions

        """
        _ = [s.forget() for s in solutions]

    def _check_validity(self, verbose=True):
        """
        Check if the dataframe is a valid HGC dataframe

        Notes
        -----
        Checks are:
            1. Are there any columns names in the recognized parameter set?
            2. Are there no strings in the recognized columns (except '<' and '>')?
            3. Are there negative concentrations in the recognized columns?

        """
        obj = self._obj
        if verbose:
            logging.info("Checking validity of DataFrame for HGC...")
        # Define allowed columns that contain concentration values
        allowed_concentration_columns = (list(constants.atoms.keys()) +
                                         list(constants.ions.keys()))

        hgc_cols = self.hgc_cols
        neg_conc_cols = []
        invalid_str_cols = []

        # Check the columns for (in)valid values
        for col in hgc_cols:
            # check for only numeric values
            if obj[col].dtype in ('object', 'str'):
                if not all(obj[col].str.isnumeric()):
                    invalid_str_cols.append(col)
            # check for non-negative concentrations
            elif (col in allowed_concentration_columns) and (any(obj[col] < 0)):
                neg_conc_cols.append(col)

        is_valid = ((len(hgc_cols) > 0) and (len(neg_conc_cols) == 0) and (len(invalid_str_cols) == 0))

        if verbose:
            logging.info(f"DataFrame contains {len(hgc_cols)} HGC-columns")
            if len(hgc_cols) > 0:
                logging.info(f"Recognized HGC columns are: {','.join(hgc_cols)}")

            unused_columns = list(set(obj.columns) - set(hgc_cols))
            if unused_columns:
                logging.info(f'These columns of the dataframe are not used by HGC: {unused_columns}')

            logging.info(f"DataFrame contains {len(neg_conc_cols)} HGC-columns with negative concentrations")
            if len(neg_conc_cols) > 0:
                logging.info(f"Columns with negative concentrations are: {','.join(neg_conc_cols)}")

            logging.info(f"DataFrame contains {len(invalid_str_cols)} HGC-columns with invalid values")
            if len(invalid_str_cols) > 0:
                logging.info(f"Columns with invalid strings are: {','.join(invalid_str_cols)}. Only '<' and '>' and numeric values are allowed.")

            if is_valid:
                logging.info("DataFrame is valid. But might still require consolidation of lab or field data (use df.hgc.consolidate) or replace NaN (use pandas fillna function)")
            else:
                logging.info("DataFrame is not HGC valid. Use the 'make_valid' method to automatically resolve issues")

        return is_valid

    @property
    def allowed_hgc_columns(self):
        """  Returns allowed columns of the hgc `SamplesFrame`"""
        return (list(constants.atoms.keys()) +
                               list(constants.ions.keys()) +
                               list(constants.properties.keys()))
    @property
    def hgc_cols(self):
        """ Return the columns that are used by hgc """
        return [item for item in self.allowed_hgc_columns if item in self._obj.columns]


    @property
    def is_valid(self):
        """ returns a boolean indicating that the columns used by hgc have
        valid values """
        is_valid = self._check_validity(verbose=False)
        return is_valid

    def _make_input_df(self, cols_req, nan_allowed=True):
        """
        Make input DataFrame for calculations. This DataFrame contains columns for each required parameter,
        which is 0 in case the parameter is not present in original HGC frame. It also
        replaces all NaN with 0 *if* NaN are allowed as indicated with the nan_allowed argument.
        """
        if not self.is_valid:
            raise ValueError("Method can only be used on validated HGC frames, use 'make_valid' to validate")


        df_in = pd.DataFrame(columns=cols_req)
        for col_req in cols_req:
            if col_req in self._obj:
                df_in[col_req] = self._obj[col_req]
            else:
                logging.info(f"Column {col_req} is not present in DataFrame, assuming concentration 0 for this compound for now.")

        isna_cols = df_in.columns[df_in.isna().any(axis=0)]
        if (not isna_cols.empty) and (not nan_allowed):
            raise ValueError(f"Column(s) {isna_cols.values} is missing or contain(s) NaN values. Add the column, or replace NaN-values with real numbers (using e.g. pandas fillna method) to be able to determine the BEX.")
        else:
            df_in = df_in.fillna(0.0)

        return df_in


    def _replace_detection_lim(self, rule="half"):
        """
        Substitute detection limits according to one of the available
        rules. Cells that contain for example '<0.3' or '> 0.3' will be replaced
        with 0.15 and 0.45 respectively (in case of rule "half").

        Parameters
        ----------
        rule : {'half', 'on', 'zero'}, default 'half'
            Rule "half" replaces cells *below* detection limit with half of the value of the detection limit;
                   and replaces cells *above* upper detection limit with 1.5 of that detection limit's value.
            Rule "at" replaces detection limit cells with the exact value of the detection limit.
            Rule "zero" replaces below detection limit cells with zero; values above the detection limit set at detection limit.
        """
        rule = str(rule)
        for col in self.hgc_cols:
            if self._obj[col].dtype in ('object', 'str'):
                is_below_dl = self._obj[col].str.contains(pat=r'^[<]\s*\d').fillna(False)
                is_above_dl = self._obj[col].str.contains(pat=r'^[>]\s*\d').fillna(False)
                lower_detection_limit = self._obj.loc[is_below_dl, col].str.extract(r'(\d+)').astype(np.float64).values
                upper_detection_limit = self._obj.loc[is_above_dl, col].str.extract(r'(\d+)').astype(np.float64).values

                if rule.lower() == 'half':
                    logging.info("Replace values below detection limit with (detection limit) / 2.")
                    self._obj.loc[is_below_dl, col] = lower_detection_limit / 2.
                    logging.info("Replace values above detection limit with 1.5 * (detection limit).")
                    self._obj.loc[is_above_dl, col] = 1.5 * upper_detection_limit
                elif rule.lower() == 'on':
                    logging.info("Replace values above and below detection limit with detection limit.")
                    self._obj.loc[is_below_dl, col] = lower_detection_limit
                    self._obj.loc[is_above_dl, col] = upper_detection_limit
                elif str(rule).lower() in ['zero', '0']:
                    logging.info("Replace values below detection limit with zero, above detection limit with (upper) detection limit.")
                    self._obj.loc[is_below_dl, col] = 0
                    self._obj.loc[is_above_dl, col] = upper_detection_limit
                else:
                    raise ValueError("Invalid rule. Allowed rules are half, on and zero.")

    def _replace_negative_concentrations(self):
        """
        Replace any negative concentrations with 0.
        """
        # Get all columns that represent chemical compounds
        # Replace negatives with 0
        for col in self.hgc_cols:
            self._obj.loc[self._obj[col] < 0, col] = 0


    def _cast_datatypes(self):
        """
        Convert all HGC-columns to their correct data type.
        """
        for col in self.hgc_cols:
            if self._obj[col].dtype in ('object', 'str'):
                self._obj[col] = pd.to_numeric(self._obj[col], errors='coerce')


    def consolidate(self, use_ph='field', use_ec='lab', use_so4='ic', use_o2='field',
                    use_temp='field', use_alkalinity='alkalinity',
                    merge_on_na=False, inplace=True):
        """
        Consolidate parameters measured with different methods to one single parameter.

        Parameters such as EC and pH are frequently measured both in the lab and field,
        and SO4 and PO4 are frequently measured both by IC and ICP-OES. Normally we prefer the
        field data for EC and pH, but ill calibrated sensors or tough field circumstances may
        prevent these readings to be superior to the lab measurement. This method allows for quick
        selection of the preferred measurement method for each parameter and select that for further analysis.

        For each consolidated parameter HGC adds a new column that is either filled with the lab measurements or the field
        measurements. It is also possible to fill it with the preferred method, and fill remaining NaN's with
        measurements gathered with the other possible method.

        Parameters
        ----------
        use_ph : {'lab', 'field', None}, default 'field'
            Which pH to use? Ignored if None.
        use_ec : {'lab', 'field', None}, default 'lab'
            Which EC to use?
        use_so4 : {'ic', 'field', None}, default 'ic'
            Which SO4 to use?
        use_o2 : {'lab', 'field', None}, default 'field'
            Which O2 to use?
        use_alkalinity: str, default 'alkalinity'
            name of the column to use for alkalinity
        merge_on_na : bool, default False
            Fill NaN's from one measurement method with measurements from other method.
        inplace : bool, default True
            Modify `SamplesFrame` in place. inplace=False is not implemented (yet)


        Raises
        ------
            ValueError: if one of the `use_` parameters is set to a column that is not in the dataframe
                        *or* if one of the default parameters is not in the dataframe while it is not
                        set to None.
        """
        if not self.is_valid:
            raise ValueError("Method can only be used on validated HGC frames, use 'make_valid' to validate")

        if inplace is False:
            raise NotImplementedError('inplace=False is not (yet) implemented.')

        param_mapping = {
            'ph': use_ph,
            'ec': use_ec,
            'SO4': use_so4,
            'O2': use_o2,
            'temp': use_temp,
        }
        if not (use_alkalinity in ['alkalinity', None]):
            try:
                self._obj['alkalinity'] = self._obj[use_alkalinity]
                self._obj.drop(columns=[use_alkalinity], inplace=True)
            except KeyError:
                raise ValueError(f"Invalid value for argument 'use_alkalinity': " +
                                f"{use_alkalinity}. It is not a column name of " +
                                f"the dataframe")


        for param, method in param_mapping.items():
            if not method:
                # user did not specify source, ignore
                continue

            if not isinstance(method, str):
                raise ValueError(f"Invalid method {method} for parameter {param}. Arg should be a string.")

            if param in self._obj.columns:
                logging.info(f"Parameter {param} already present in DataFrame, ignoring. Remove column manually to enable consolidation.")
                continue

            source = f"{param}_{method}"
            if source in self._obj.columns:
                source_val = self._obj[source]
                if any(np.isnan(source_val)):
                    raise ValueError('Nan value for column {source}')

                self._obj[param] = np.NaN
                self._obj[param].fillna(source_val, inplace=True)


                if merge_on_na:
                    raise NotImplementedError('merge_on_na is True is not implemented (yet).')

                # Drop source columns
                suffixes = ('_field', '_lab', '_ic')
                cols = [param + suffix for suffix in suffixes]
                self._obj.drop(columns=cols, inplace=True, errors='ignore')

            else:
                raise ValueError(f"Column {source} not present in DataFrame. Use " +
                                 f"use_{param.lower()}=None to explicitly ignore consolidating " +
                                 f"this column.")


    def get_bex(self, watertype="G", inplace=True):
        """
        Get Base Exchange Index (meq/L). By default this is the BEX without dolomite.

        Parameters
        ----------
        watertype : {'G', 'P'}, default 'G'
            Watertype (Groundwater or Precipitation)
        inplace: bool, optional, default True
                whether the saturation index should be added to the `pd.DataFrame` (inplace=True)
                as column `si_<mineral_name>` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` or `pd.Series` with base exchange index for each row in SamplesFrame
            if `inplace=False`.
        """
        cols_req = ('Na', 'K', 'Mg', 'Cl')
        df = self._make_input_df(cols_req, nan_allowed=False)

        df_out = pd.DataFrame()

        #TODO: calculate alphas on the fly from SMOW constants
        alpha_na = 0.556425145165362 # ratio of Na to Cl in SMOW
        alpha_k = 0.0206 # ratio of K to Cl in SMOW
        alpha_mg = 0.0667508204998738 # ratio of Mg to Cl in SMOW

        only_p_and_t = True

        if watertype == "P" and only_p_and_t:
            df_out['Na_nonmarine'] = df['Na'] - 1.7972 * alpha_na*df['Na']
            df_out['K_nonmarine'] = df['K'] - 1.7972 * alpha_k*df['Na']
            df_out['Mg_nonmarine'] = df['Mg'] - 1.7972 * alpha_mg*df['Na']
        else:
            df_out['Na_nonmarine'] = df['Na'] - alpha_na*df['Cl']
            df_out['K_nonmarine'] = df['K'] - alpha_k*df['Cl']
            df_out['Mg_nonmarine'] = df['Mg'] - alpha_mg*df['Cl']

        df_out['bex'] = df_out['Na_nonmarine']/22.99 + df_out['K_nonmarine']/39.098 + df_out['Mg_nonmarine']/12.153

        if inplace:
            self._obj['bex'] = df_out['bex']
        else:
            return df_out['bex']


    @requires_ph
    def get_ratios(self, inplace=True):
        """
        Calculate common hydrochemical ratios, will ignore any ratios
        in case their constituents are not present in the `SamplesFrame`.
        requires pH to determine the sum of anions

        Notes
        -----

        It is assumed that only |HCO3| contributes to
        the alkalinity.

        HGC will attempt to calculate the following ratios:

         * Cl/Br (molar ratio)
         * Cl/Na (molar ratio)
         * Cl/Mg (molar ratio)
         * Ca/Sr (molar ratio)
         * Fe/Mn (molar ratio)
         * HCO3/Ca (molar ratio)
         * 2H/18O (molar ratio)
         * SUVA: UVA254/DOC
         * HCO3/Sum of anions (molar ratio)
         * HCO3/Sum of Ca and Mg (molar ratio)
         * MONC (Mean Oxidation Number of Carbon)
         * COD/DOC

        Parameters
        ----------
        inplace: bool, optional, default True
                whether the saturation index should be added to the `pd.DataFrame` (inplace=True)
                as column `<numerator>_to_<denominator>` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.DataFrame or None
            Returns None if `inplace=True` and `pd.DataFrame` with the different ratios as column, for each row in
            `SamplesFrame` if `inplace=False`.
        """
        if not self.is_valid:
            raise ValueError("Method can only be used on validated HGC frames, use 'make_valid' to validate")

        df_ratios = pd.DataFrame()

        ratios = {
            'cl_to_br': ['Cl', 'Br'],
            'cl_to_na': ['Cl', 'Na'],
            'ca_to_mg': ['Ca', 'Mg'],
            'ca_to_sr': ['Ca', 'Sr'],
            'fe_to_mn': ['Fe', 'Mn'],
            'hco3_to_ca': ['alkalinity', 'Ca'],
            '2h_to_18o': ['2H', '18O'],
            'suva': ['uva254', 'doc'], # TODO check if this and following lines is used
            'hco3_to_sum_anions': ['alkalinity',self.__SUM_ANIONS_COLUMN],
            'hco3_to_ca_and_mg': ['alkalinity', 'Ca', 'Mg'],
            'monc': ['cod', 'Fe', 'NO2', 'doc'],
            'cod_to_doc': ['cod', 'Fe', 'NO2', 'doc']
        }
        # add sum_anions column to the df
        try:
            # this is only necessary if the ratio to anions is required, but
            # calculate it always to make the part below simpler
            self._obj[self.__SUM_ANIONS_COLUMN]
        except KeyError:
            self.get_sum_anions(inplace=True)

        if 'alkalinity' in self._obj.columns:
            hco3 = self._obj['alkalinity'] / mw('HCO3')

        for ratio, constituents in ratios.items():
            has_cols = [const in self._obj.columns for const in constituents]
            if all(has_cols):
                if ratio == 'hco3_to_sum_anions':
                    df_ratios[ratio] = hco3  / self._obj[self.__SUM_ANIONS_COLUMN]
                elif ratio == 'hco3_to_ca':
                    df_ratios[ratio] = hco3 / (self._obj['Ca'] / mw('Ca'))
                elif ratio == 'hco3_to_ca_and_mg':
                    df_ratios[ratio] = hco3 / ((self._obj['Ca'] / mw('Ca')) + ( self._obj['Mg'] / mw('Mg') ))
                elif ratio == 'monc':
                    df_ratios[ratio] = 4 - 1.5 * (self._obj['cod'] - 0.143 * self._obj['Fe'] - 0.348 * self._obj['NO2']) / (3.95 * self._obj['doc'])
                elif ratio == 'cod_to_doc':
                    # TODO: CHEKC BRACKETS, abstract COD_O out of equation
                    df_ratios[ratio] = ((0.2532 * self._obj['cod'] - 0.143 * self._obj['Fe'] - 0.348 * self._obj['NO2']) / 32) / (self._obj['doc'] / 12)
                else:
                    numerator = self._obj[constituents[0]] / mw(constituents[0])  # to mol
                    denominator = self._obj[constituents[1]] / mw(constituents[1])  # to mol

                    df_ratios[ratio] = numerator / denominator
            else:
                missing_cols = [i for (i, v) in zip(constituents, has_cols) if not v]
                logging.info(f"Cannot calculate ratio {ratio} since columns {','.join(missing_cols)} are not present.")

        if inplace:
            logging.info(f'Added columns {list(df_ratios.columns)}')
            self._obj[df_ratios.columns] = df_ratios
        else:
            return df_ratios


    @requires_ph
    def get_stuyfzand_water_type(self, inplace=True):
        """
        Get Stuyfzand water type. This water type classification contains
        5 components: Salinity, Alkalinity, Dominant Cation, Dominant Anion and Base Exchange Index.
        This results in a classification such as for example 'F3CaMix+'.

        It is assumed that only |HCO3| contributes to
        the alkalinity.

        See http://www.hydrology-amsterdam.nl/valorisation/HGCmanual_v2_1.pdf chapter 5 for the definitions.

        Parameters
        ----------
        inplace: bool, optional, default True
                whether the saturation index should be added to the `pd.DataFrame` (inplace=True)
                as column `<numerator>_to_<denominator>` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series
            Series with Stuyfzand water type of each row in original `SamplesFrame`.
        """
        if not self.is_valid:
            raise ValueError("Method can only be used on validated HGC frames, use 'make_valid' to validate")

        # Create input dataframe containing all required columns
        # Inherit column values from HGC frame, assume 0 if column
        # is not present
        cols_req = ('Al', 'Ba', 'Br', 'Ca', 'Cl', 'Co', 'Cu', 'doc', 'F', 'Fe', 'alkalinity', 'K', 'Li', 'Mg', 'Mn', 'Na', 'Ni', 'NH4', 'NO2', 'NO3', 'Pb', 'PO4', 'ph', 'SO4', 'Sr', 'Zn')
        df_in = self._make_input_df(cols_req)
        df_out = pd.DataFrame(index=df_in.index)

        # Salinity
        df_out['swt_s'] = 'G'
        df_out.loc[df_in['Cl'] > 5, 'swt_s'] = 'g'
        df_out.loc[df_in['Cl'] > 30, 'swt_s'] = 'F'
        df_out.loc[df_in['Cl'] > 150, 'swt_s'] = 'f'
        df_out.loc[df_in['Cl'] > 300, 'swt_s'] = 'B'
        df_out.loc[df_in['Cl'] > 1000, 'swt_s'] = 'b'
        df_out.loc[df_in['Cl'] > 10000, 'swt_s'] = 'S'
        df_out.loc[df_in['Cl'] > 20000, 'swt_s'] = 'H'

        #Alkalinity
        df_out['swt_a'] = '*'
        df_out.loc[df_in['alkalinity'] > 31, 'swt_a'] = '0'
        df_out.loc[df_in['alkalinity'] > 61, 'swt_a'] = '1'
        df_out.loc[df_in['alkalinity'] > 122, 'swt_a'] = '2'
        df_out.loc[df_in['alkalinity'] > 244, 'swt_a'] = '3'
        df_out.loc[df_in['alkalinity'] > 488, 'swt_a'] = '4'
        df_out.loc[df_in['alkalinity'] > 976, 'swt_a'] = '5'
        df_out.loc[df_in['alkalinity'] > 1953, 'swt_a'] = '6'
        df_out.loc[df_in['alkalinity'] > 3905, 'swt_a'] = '7'

        #Dominant cation
        s_sum_cations = self.get_sum_cations(inplace=False)

        df_out['swt_domcat'] = self.get_dominant_cations(inplace=False)

        # Dominant anion
        df_out['swt_doman'] = self.get_dominant_anions(inplace=False)

        # Base Exchange Index
        s_bex = self.get_bex(inplace=False)
        s_sum_anions = self.get_sum_anions(inplace=False)
        cl_mmol = df_in.Cl/mw('Cl')

        threshold1 = 0.5 + 0.02*cl_mmol
        threshold2 = -0.5-0.02*cl_mmol
        is_plus = (s_bex > threshold1) & (s_bex > 1.5*(s_sum_cations-s_sum_anions))

        is_minus = ~is_plus & (s_bex < threshold2) & (s_bex < 1.5*(s_sum_cations-s_sum_anions))

        is_neutral = (~is_plus & ~is_minus &
                      (s_bex > threshold2) & (s_bex < threshold1) &
                      ((s_sum_cations == s_sum_anions) |
                       ((abs(s_bex + threshold1*(s_sum_cations-s_sum_anions))/abs(s_sum_cations-s_sum_anions))
                        > abs(1.5*(s_sum_cations-s_sum_anions)))
                       )
                      )

        is_none = ~is_plus & ~is_minus & ~is_neutral


        df_out.loc[is_plus, 'swt_bex'] = '+'
        df_out.loc[is_minus, 'swt_bex'] = '-'
        df_out.loc[is_neutral, 'swt_bex'] = 'o'
        df_out.loc[is_none, 'swt_bex'] = ''

        #Putting it all together
        df_out['swt'] = df_out['swt_s'].str.cat(df_out[['swt_a', 'swt_domcat', 'swt_doman', 'swt_bex']])

        if inplace:
            logging.info(f'Added the column water_type.')
            self._obj['water_type'] = df_out['swt']
        else:
            return df_out['swt']

    @requires_ph
    def get_dominant_cations(self, inplace=True):
        """  calculates the dominant cations of the SamplesFrame as used by the Stuyfzand water type classification (
        See: http://www.hydrology-amsterdam.nl/valorisation/HGCmanual_v2_1.pdf chapter 5 for the definitions.)

        Parameters
        -----------
        inplace: bool, optional, default True
                whether the dominant cations should be added to the `SamplesFrame` (inplace=True)
                as column `dominant_cation` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` or `pd.Series` with dominant cation for each row in `SamplesFrame`
            if `inplace=False`.
        """
        s_sum_cations = self.get_sum_cations(inplace=False)

        cols_req = ('ph', 'Na', 'K', 'Ca', 'Mg', 'Fe', 'Mn', 'NH4', 'Al', 'Ba', 'Co', 'Cu', 'Li', 'Ni', 'Pb', 'Sr', 'Zn')
        df_in = self._make_input_df(cols_req)

        na_mmol = df_in.Na/mw('Na')
        k_mmol = df_in.K/mw('K')
        nh4_mmol = df_in.NH4/(mw('N')+4*mw('H'))
        ca_mmol = df_in.Ca/mw('Ca')
        mg_mmol = df_in.Mg/mw('Mg')
        fe_mmol = df_in.Fe/mw('Fe')
        mn_mmol = df_in.Mn/mw('Mn')
        h_mmol = (10**-df_in.ph) / 1000  # ph -> mol/L -> mmol/L
        al_mmol = 1000. * df_in.Al/mw('Al')  # ug/L ->mg/L -> mmol/L

        # - Na, K, NH4
        # select rows that do not have Na, K or NH4 as dominant cation
        is_no_domcat_na_nh4_k = (na_mmol + k_mmol + nh4_mmol) < (s_sum_cations/2)

        is_domcat_nh4 = ~is_no_domcat_na_nh4_k & (nh4_mmol > (na_mmol + k_mmol))

        is_domcat_na = ~is_no_domcat_na_nh4_k & ~is_domcat_nh4 & (na_mmol > k_mmol)

        is_domcat_k = ~is_no_domcat_na_nh4_k & ~is_domcat_nh4 & ~is_domcat_na

        # abbreviation
        is_domcat_na_nh4_k = is_domcat_na | is_domcat_nh4 | is_domcat_k

        # - Ca, Mg
        is_domcat_ca_mg = (
            # not na or nh4 or k dominant
            ~is_domcat_na_nh4_k & (
                # should be any of Ca or Mg available
                ((ca_mmol > 0) | (mg_mmol > 0)) |
                # should be more of Ca or Mg then sum of H, Fe, Al, Mn
                # (compensated for charge)
                (2*ca_mmol+2*mg_mmol < h_mmol+3*al_mmol+2*fe_mmol+2*mn_mmol)))

        is_domcat_ca = is_domcat_ca_mg & (ca_mmol >= mg_mmol)
        is_domcat_mg = is_domcat_ca_mg & (ca_mmol < mg_mmol)

        # - H, Al, Fe, Mn
        # IF(IF(h_mmol+3*IF(al_mmol)>2*(fe_mol+mn_mol),IF(h_mmol>3*al_mmol,"H","Al"),IF(fe_mol>mn_mol,"Fe","Mn")))
        is_domcat_fe_mn_al_h = (
            # not na, nh4, k, ca or Mg dominant
            ~is_domcat_na_nh4_k & ~is_domcat_ca & ~is_domcat_mg & (
                # should be any of Fe, Mn, Al or H available
                (fe_mmol > 0) | (mn_mmol > 0) | (h_mmol > 0) | (al_mmol > 0)  # |
                # # should be more of Ca or Mg then sum of H, Fe, Al, Mn
                # # (compensated for charge)
                # (2*ca_mmol+2*mg_mmol < h_mmol+3*al_mmol+2*fe_mmol+2*mn_mmol)
            )
        )

        is_domcat_h_al=is_domcat_fe_mn_al_h & ((h_mmol + 3*al_mmol) > (2*fe_mmol + 2*mn_mmol))
        is_domcat_h = is_domcat_h_al & (h_mmol > al_mmol)
        is_domcat_al = is_domcat_h_al & (al_mmol > h_mmol)

        is_domcat_fe_mn = is_domcat_fe_mn_al_h & ~is_domcat_h_al
        is_domcat_fe = is_domcat_fe_mn & (fe_mmol > mn_mmol)
        is_domcat_mn = is_domcat_fe_mn & (mn_mmol > fe_mmol)

        sr_dominant_cation = pd.Series(index=df_in.index, dtype='object')
        sr_dominant_cation[:] = ""
        sr_dominant_cation[is_domcat_nh4] = "NH4"
        sr_dominant_cation[is_domcat_na] = "Na"
        sr_dominant_cation[is_domcat_k] = "K"
        sr_dominant_cation[is_domcat_ca] = 'Ca'
        sr_dominant_cation[is_domcat_mg] = 'Mg'
        sr_dominant_cation[is_domcat_fe] = 'Fe'
        sr_dominant_cation[is_domcat_mn] = 'Mn'
        sr_dominant_cation[is_domcat_al] = 'Al'
        sr_dominant_cation[is_domcat_h] = 'H'

        if inplace:
            logging.info("Dominant cation is added to SamplesFrame as column 'dominant_cation' to the DataFrame")
            self._obj['dominant_cation'] = sr_dominant_cation
        else:
            return sr_dominant_cation


    def get_dominant_anions(self, inplace=True):
        """  calculates the dominant anion of each row in the SamplesFrame as used by the Stuyfzand water type classification (
        See: http://www.hydrology-amsterdam.nl/valorisation/HGCmanual_v2_1.pdf chapter 5 for the definitions.)

        Parameters
        ----------
        inplace: bool, optional, default True
                whether the dominant anion should be added to the `pd.DataFrame`
                as column `dominant_anion` (inplace=True) or returned as a
                `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` or `pd.Series` with dominant anion for each row in `SamplesFrame`
            if `inplace=False`.
        """
        s_sum_anions = self.get_sum_anions(inplace=False)
        df_in = self._obj
        cl_mmol = df_in.Cl/mw('Cl')
        hco3_mmol = df_in.alkalinity/(mw('H') + mw('C') + 3*mw('O'))
        no3_mmol = df_in.NO3/(mw('N') + 3*mw('O'))
        so4_mmol = df_in.SO4/(mw('S') + 4*mw('O'))

        # TODO: consider renaming doman to dom_an or dom_anion
        is_doman_cl = (cl_mmol > s_sum_anions/2)
        sr_dominant_anions = pd.Series(index=self._obj.index, dtype='object')
        sr_dominant_anions[is_doman_cl] = "Cl"

        is_doman_hco3 = ~is_doman_cl & (hco3_mmol > s_sum_anions/2)
        sr_dominant_anions[is_doman_hco3] = "HCO3"

        is_doman_so4_or_no3 = ~is_doman_cl & ~is_doman_hco3 & (2*so4_mmol + no3_mmol > s_sum_anions/2)
        is_doman_so4 = (2*so4_mmol > no3_mmol)
        sr_dominant_anions[is_doman_so4_or_no3 & is_doman_so4] = "SO4"
        sr_dominant_anions[is_doman_so4_or_no3 & ~is_doman_so4] = "NO3"

        is_mix = ~is_doman_cl & ~is_doman_hco3 & ~is_doman_so4_or_no3
        sr_dominant_anions[is_mix] = "Mix"

        if inplace:
            logging.info("Dominant anion is added to SamplesFrame as column 'dominant_anion' to the DataFrame")
            self._obj['dominant_anion'] = sr_dominant_anions
        else:
            return sr_dominant_anions

    def fillna_concentrations(self, how="phreeqc"):
        """
        Calculate missing concentrations based on the charge balance.

        Parameters
        ----------
        how : {'phreeqc', 'analytic'}, default 'phreeqc'
            Method to compute missing concentrations.
        """
        raise NotImplementedError()

    def get_ion_balance(self, inplace=True):
        """
        Calculate the balance between anion and cations and add it as a percentage [%]
        to the column 'ion_balance' to the SamplesFrame

        Parameters
        -----------
        inplace: bool, optional, default True
                whether the ion balance should be added to the `SamplesFrame` (inplace=True)
                as column `ion_balance` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` or `pd.Series` with ion balance for each row in `SamplesFrame`
            if `inplace=False`.
        """
        anions = abs(self.get_sum_anions(inplace=False))
        cations = abs(self.get_sum_cations(inplace=False))
        ion_balance = 100 * (cations - anions) / (cations + anions)
        if inplace:
            logging.info("Charge balance of ions is added to the column ion_balance to the DataFrame")
            self._obj['ion_balance'] = ion_balance
        else:
            return ion_balance


    def fillna_ec(self, use_phreeqc=True):
        """
        Calculate missing Electrical Conductivity measurements using
        known anions and cations.
        """
        if use_phreeqc:
            # use get_specific_conductance method on
            # all N/A rows of EC columns
            raise NotImplementedError()
        else:
            raise NotImplementedError()

    def make_valid(self):
        """
        Try to convert the DataFrame into a valid HGC-`SamplesFrame`.
        """
        # Conduct conversions here. If they fail, raise error (e.g. when not a single valid column is present)
        # Important: order is important, first convert strings to double, then replace negative concentrations
        self._replace_detection_lim()
        self._cast_datatypes()
        self._replace_negative_concentrations()
        self._check_validity(verbose=True)


    @requires_ph
    def get_sum_anions(self, inplace=True):
        """
        Calculate sum of anions in milli equivalents (meq) according to the Stuyfzand method; i.e.
        correcting for the charge on the ions

        It is assumed that only |HCO3| contributes to
        the alkalinity.

        Parameters
        ----------
        inplace: bool, optional, default True
                whether the sum of anions should be added to the `pd.DataFrame` (inplace=True)
                as column `sum_anions` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` or `pd.Series` with sum of anions for each row in `SamplesFrame`
            if `inplace=False`.
        """
        cols_req = ('Br', 'Cl', 'doc', 'F', 'alkalinity', 'NO2', 'NO3', 'PO4', 'SO4', 'ph')
        df_in = self._make_input_df(cols_req)
        s_sum_anions = pd.Series(index=df_in.index,dtype='float64')

        sum_ions = (df_in['Cl']/mw('Cl') + df_in['SO4']/(mw('SO4')/2) +
                    df_in['alkalinity']/mw('HCO3') + df_in['NO3']/mw('NO3') +
                    df_in['NO2']/mw('NO2') + df_in['F']/mw('F') +
                    df_in['Br']/(mw('Br')*1000.) +
                    (df_in['PO4']/mw('PO4')) / (1 + 10**(df_in['ph']-7.21)) # correcting for the charge *and* protonation of (H)PO3 with different pH
                   )

        k_org = 10**(0.039*df_in['ph']**2 - 0.9*df_in['ph']-0.96) # HGC manual equation 3.5
        a_org = k_org * df_in['doc'] / (100.*k_org + (10.**-df_in['ph'])/10.) # HGC manual equation 3.4A
        is_add_a_org = (a_org > df_in['alkalinity']/61.02)

        s_sum_anions.loc[is_add_a_org] = sum_ions + a_org
        s_sum_anions.loc[~is_add_a_org] = sum_ions

        if inplace:
            self._obj[self.__SUM_ANIONS_COLUMN] = s_sum_anions
        else:
            return s_sum_anions

    @requires_ph
    def get_sum_cations(self, inplace=True):
        """
        Calculate sum of cations according to the Stuyfzand method.

        Parameters
        ----------
        inplace: bool, optional, default True
                whether the sum of cations should be added to the `pd.DataFrame` (inplace=True)
                as column `sum_cations` or returned as a `pd.Series` (inplace=False).

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` or `pd.Series` with sum of cations for each row in `SamplesFrame`
            if `inplace=False`.
        """
        cols_req = ('ph', 'Na', 'K', 'Ca', 'Mg', 'Fe', 'Mn', 'NH4', 'Al', 'Ba', 'Co', 'Cu', 'Li', 'Ni', 'Pb', 'Sr', 'Zn')
        df_in = self._make_input_df(cols_req)

        s_sum_cations = 10**-(df_in['ph']-3) + \
                    df_in['Na']/22.99 + \
                    df_in['K']/39.1 + \
                    df_in['Ca']/20.04 + \
                    df_in['Mg']/12.156 + \
                    df_in['Fe']/(55.847/2) + \
                    df_in['Mn']/(54.938/2) + \
                    df_in['NH4']/18.04 + \
                    df_in['Al']/(26982/3) + \
                    df_in['Ba']/137327 + \
                    df_in['Co']/58933 + \
                    df_in['Cu']/(63546/2) + \
                    df_in['Li']/6941 + \
                    df_in['Ni']/58693 + \
                    df_in['Pb']/207200 + \
                    df_in['Sr']/87620 + \
                    df_in['Zn']/65380

        if inplace:
            self._obj['sum_cations'] = s_sum_cations
        else:
            return s_sum_cations


    @requires_ph
    def select_phreeq_columns(self):
        """
        Returns the columns from the DataFrame that might be used
        by PhreeqPython.

        Returns
        -------
        list
            Usable PhreeqPython columns
        """
        df = self._obj

        bicarbonate_in_columns = any([('hco' in _c.lower()) or
                                      ('bicarbona' in _c.lower())
                                      for _c in df.columns])
        alkalinity_in_columns = any(['alkalinity' in _c.lower()
                                     for _c in df.columns])
        if bicarbonate_in_columns:
            if alkalinity_in_columns:
                logging.warning('Warning: both bicarbonate (or hco3) and alkalinity are ' +
                'defined as columns. Note that only alkalinity column is used')
            else:
                logging.warning('Warning: bicarbonate (or hco3) is found, but no alkalinity ' +
                'is defined as columns. Note that only alkalinity column are used')


        atom_columns = set(self._valid_atoms).intersection(df.columns)
        ion_columns = set(self._valid_ions).intersection(df.columns)
        prop_columns = set(self._valid_properties).intersection(df.columns)
        phreeq_columns = list(atom_columns.union(ion_columns).union(prop_columns))

        nitrogen_cols = set(phreeq_columns).intersection({'NO2', 'NO3', 'N', 'N_tot_k'})
        phosphor_cols = set(phreeq_columns).intersection({'PO4', 'P', 'P_ortho', 'PO4_total'})

        # check whether temp column exists
        if 'temp' not in phreeq_columns:
            raise ValueError('The required column temp is missing in the dataframe. ' +
                             'Add a column temp manually or consolidate temp_lab or temp_field ' +
                             'to temp by running the method DataFrame.hgc.consolidate().')
        if 'doc' in phreeq_columns:
            logging.info('DOC column found in samples frame while using phreeqc as backend; influence of DOC on any value calculated using phreeqc is ignored.')

        if len(nitrogen_cols) > 1:
            # check if nitrogen is defined in more than one column (per sample)
            duplicate_nitrogen = df[nitrogen_cols].apply(lambda x: sum(x > 0) > 1, axis=1)

            if sum(duplicate_nitrogen) > 0:
                logging.info('Some rows have more than one column defining N. ' +
                             'Choose N over NO2 over NO3')

            for index, row in df.loc[duplicate_nitrogen].iterrows():
                for col in ['N', 'NO2', 'NO3']:
                    if col in nitrogen_cols:
                        if row[col] > 0:
                            df.loc[index, list(nitrogen_cols-{col})] = 0.
                            break

        if len(phosphor_cols) > 1:
            # check if phosphor is defined in more than one column (per sample)
            duplicate_phosphor = df[phosphor_cols].apply(lambda x: sum(x > 0) > 1, axis=1)

            if sum(duplicate_phosphor) > 0:
                logging.info('Some rows have more than one column defining P. Choose P over PO4')

            for index, row in df.loc[duplicate_phosphor].iterrows():
                for col in ['P', 'PO4']:
                    if col in phosphor_cols:
                        if row[col] > 0:
                            df.loc[index, list(phosphor_cols-{col})] = 0.
                            break


        return phreeq_columns

    def get_phreeqpython_solutions(self, equilibrate_with='none', inplace=True):
        """
        Return a series of `phreeqpython solutions <https://github.com/Vitens/phreeqpython>`_ derived from the (row)data in the `SamplesFrame`.

        Parameters
        ----------
        equilibrate_with : str, default 'none'
            Ion to add for achieving charge equilibrium in the solutions.
        inplace : bool, default True
            Whether the result is returned as a `pd.Series` or is added to the `pd.DataFrame`
            as column `pp_solutions`.

        Returns
        -------
        pandas.Series or None
            Returns None if `inplace=True` and `pd.Series` with `PhreeqPython.Solution` instances for every row in
            `SamplesFrame` if `inplace=False`.
        """
        # `None` is also a valid argument and is translated to the strin `'none'`
        if equilibrate_with is None:
            equilibrate_with = 'none'

        pp = self._pp
        df = self._obj.copy()

        phreeq_cols = self.select_phreeq_columns()

        solutions = pd.Series(index=df.index, dtype='object')
        for index, row in df[phreeq_cols].iterrows():
            _sol = {'units': 'mg/l'}
            for col in row.index:

                try:
                    atom = self._valid_atoms[col]
                    phreeq_name = atom.feature
                    phreeq_as = ''
                    phreeq_unit = atom.unit
                    value = row[col]
                except KeyError:
                    try:
                        ion = self._valid_ions[col]
                        phreeq_name = ion.phreeq_name
                        phreeq_as = ion.phreeq_concentration_as
                        phreeq_unit = ion.unit
                        if phreeq_as is None:
                            phreeq_as = ''
                        value = row[col]
                    except KeyError:
                        property_ = self._valid_properties[col]
                        phreeq_name = property_.phreeq_name
                        phreeq_as = ''
                        phreeq_unit = ''
                        value = row[col]

                # phreeqc cannot cope with μ, so replace with u
                phreeq_unit = phreeq_unit.replace('μ', 'u')

                if (value > 0) and (phreeq_name is not None):
                    # Phreeq_name is None indicates that the atom/ion/property
                    # is not a valid property for a phreeqc simulation and the key
                    # should therefore not be added to the phreeqpython solutions
                    phreeq_name = phreeq_name.strip()
                    phreeq_unit = phreeq_unit.strip()
                    phreeq_as = phreeq_as.strip()
                    _sol[phreeq_name] = f"{value} {phreeq_unit} {phreeq_as}"

            if not (equilibrate_with.lower() in ['none', 'auto']):
                try:
                    # append the keyword charge to the compound that is used to charge balance
                    _sol[equilibrate_with] = _sol[equilibrate_with] + ' charge'
                except KeyError:
                    logging.info(f'{equilibrate_with} not found in solution while it is selected to balance charge with. Starts initial guess with 20 mg/L to balance charge.')
                    _sol[equilibrate_with] = '20. mg/L charge'
            elif equilibrate_with.lower() == 'auto':
                try:
                    # append the keyword charge to the compound that is used to charge balance
                    _sol['Na'] = _sol['Na'] + ' charge'
                except KeyError:
                    logging.info(f'Na is not found in solution while it is automatically selected to balance charge with. Starts initial guess with 20 mg/L Na to balance charge.')
                    _sol[equilibrate_with] = '20. mg/L charge'

            try:
                solutions[index] = pp.add_solution(_sol)
            except Exception as error:
                if equilibrate_with.lower() == 'auto':
                    _sol['Na'] = _sol['Na'].replace(' charge', '')
                    _sol['Cl'] = _sol['Cl'] + ' charge'
                    try:
                        logging.info(f"initializing solution with charge balancing with Na failed. Now trying to initialize solution by" +
                                     " charge balancing with Cl.")
                        solutions[index] = pp.add_solution(_sol)
                    except Exception as error:
                        logging.info(error)
                        raise ValueError(f'Something went wrong with the phreeqc calculation with index {index} from the DataFrame. PHREEQC returned: {error}. Charge balancing with either Na or Cl failed.')
                else:
                    logging.info(error)
                    raise ValueError(f'Something went wrong with the phreeqc calculation with index {index} from the DataFrame. PHREEQC returned: {error}. ' +
                                     'Possibly charge balance could (sufficiently) reached.')

        return_series = pd.Series(solutions, index=self._obj.index)
        if inplace:
            self._obj['pp_solutions'] = return_series
        else:
            # return the solutions as pandas series with the same index as the source dataframe
            return return_series

    def get_saturation_index(self, mineral_or_gas, use_phreeqc=True, inplace=True, **kwargs):
        """ adds or returns the saturation index (SI) of a mineral or the partial pressure of a gas using phreeqc. The
            column name of the result is si_<mineral_name> in lower case (if inplace=True).

           Parameters
           ----------
           mineral_or_gas: str
                           the name of the mineral of which the SI needs to be calculated
           use_phreeqc: bool
                        whether to return use phreeqc as backend or fall back on internal hgc-routines to calculate SI
                        or partial pressure
            inplace: bool, optional, default=True
                    whether the saturation index should be added to the `pd.DataFrame` (inplace=True)
                    as column `si_<mineral_name>` or returned as a `pd.Series` (inplace=False).

            Returns
            -------
            pandas.Series or None
                Returns None if `inplace=True` and `pd.Series` with the saturation index of the mineral for each row in `SamplesFrame`
                if `inplace=False`.
        """
        if not use_phreeqc:
            raise NotImplementedError('use_phreeqc=False is not yet implemented.')


        solutions = self.get_phreeqpython_solutions(inplace=False, **kwargs)
        saturation_index = [s.si(mineral_or_gas) if s is not None else None for s in solutions]

        self._clean_up_phreeqpython_solutions(solutions)

        # return it as series with the same index as the dataframe
        name_series = 'si_'+ mineral_or_gas.lower()
        return_series = pd.Series(saturation_index, index=self._obj.index,
                                  name=name_series)
        if inplace:
            logging.info(f'Added column {name_series}')
            self._obj[name_series] = return_series
        else:
            return return_series

    def get_partial_pressure(self, gas, use_phreeqc=True, inplace=True, **kwargs):
        """ adds or returns the partial pressure of a gas using phreeqc. It is an alias for `get_saturation_index` so
            look at that method for details. gas column is pp_<gas_name>

        """
        pp_gas = self.get_saturation_index(gas, use_phreeqc, inplace=False, **kwargs)
        name_series = 'pp_'+ gas.lower()
        if inplace:
            logging.info(f'Added column {name_series}')
            self._obj[name_series] = pp_gas
        else:
            return pp_gas

    def get_specific_conductance(self, use_phreeqc=True, inplace=True, **kwargs):
        """ returns the specific conductance (sc) of a water sample using phreeqc. sc is
            also known as electric conductivity (ec) or egv measurements.

            Parameters
            ----------
            use_phreeqc: bool, optional
                    whether to return use phreeqc as backend or fall back on internal hgc-routines to calculate SI
                    or partial pressure
            inplace: bool, optional, default=True
                    whether the specific conductance should be added to the `pd.DataFrame` (inplace=True)
                    as column `sc` or returned as a `pd.Series` (inplace=False).
            **kwargs:
                     are passed to the method `get_phreeqpython_solutions`

            Returns
            -------
            pandas.Series or None
                Returns None if `inplace=True` and `pd.Series` with specific conductance for each row in `SamplesFrame`
                if `inplace=False`.
        """
        if not use_phreeqc:
            raise NotImplementedError('use_phreeqc=False is not yet implemented.')

        # create phreeqpython solutions
        solutions = self.get_phreeqpython_solutions(inplace=False, **kwargs)
        # extract sc from them
        specific_conductance = [s.sc for s in solutions]
        # clean up
        self._clean_up_phreeqpython_solutions(solutions)

        # return it as series with the same index as the dataframe
        series_name = 'sc'
        return_series = pd.Series(specific_conductance, index=self._obj.index,
                                  name=series_name)
        if inplace:
            self._obj[series_name] = return_series
        else:
            return return_series



from hgc.constants.read_write import convert_csv_to_tuples

atoms, ions, properties = convert_csv_to_tuples()

def mw(formula):
    ''' convenience function to return the molar
        weight of a molecule with `formula` '''
    try:
        if formula == 'HCO3':
            return atoms['H'].mw + atoms['C'].mw + 3*atoms['O'].mw
        elif formula == 'CO3':
            return atoms['C'].mw + 3*atoms['O'].mw
        elif formula == 'SO4':
            return atoms['S'].mw + 4*atoms['O'].mw
        elif formula == 'SO3':
            return atoms['S'].mw + 3*atoms['O'].mw
        elif formula == 'NO3':
            return atoms['N'].mw + 3*atoms['O'].mw
        elif formula == 'NO2':
            return atoms['N'].mw + 2*atoms['O'].mw
        elif formula == 'PO4':
            return atoms['P'].mw + 4*atoms['O'].mw
        else:
            return atoms[formula].mw
    except KeyError:
        raise KeyError(f'{formula} is not an element and thus not a valid key,'
                        + ' in the atoms tabel.'
                        + f'valid keys are {list(atoms.keys())}')


def units(item):
    """ returns the unit of the column of the hgc SamplesFrame """
    try:
        return atoms[item].unit
    except KeyError:
        try:
            unit_as = ions[item].phreeq_concentration_as
            if unit_as is None:
                return ions[item].unit
            else:
                return ions[item].unit + ' ' + unit_as
        except KeyError:
            try:
                return properties[item].unit
            except KeyError:
                valid_keys = (list(atoms.keys()) + list(ions.keys()) +
                              list(properties.keys()))
                raise KeyError(f'{item} is not a valid key in the atoms,'
                               + ' ions or properties tables.'
                               + f' valid keys are {valid_keys}')

def units_wt_as(item):
    try:
        return atoms[item].unit
    except KeyError:
        try:
            return ions[item].unit
        except KeyError:
            try:
                return properties[item].unit
            except KeyError:
                return None # return a None unit if the feature is excluded in HGC



""" Import csv files that contain information on the constants required by HGC.
    """
from collections import namedtuple
from pathlib import Path

import pandas as pd
from pyparsing import Group, OneOrMore, Optional, Word

PATH = Path(__file__).parent

def _formula_parser(formula, calculate_or_not, atoms):
    ''' parses the chemical formula and determine its mol weight. But only
        if `calculate_or_not` equals `calculate`, otherwise, do nothing. Use
        the molecular weight from atoms dict. '''

    if (calculate_or_not != 'calculate') or (formula in ['N_tot_k', 'PO4_ortho', 'SO4_ic', 'alkalinity']):
        return calculate_or_not

    # define some strings to use later, when describing valid lists
    # of characters for chemical symbols and numbers
    caps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    lowers = caps.lower()
    digits = "0123456789"
    brackets = "[]{}()"

    if any([_ in formula for _ in brackets]):
        raise Exception('Not implemented with brackets in formula string')

    # Version 1
    # Define grammar for a chemical formula
    # - an element is a Word, beginning with one of the characters in caps,
    #   followed by zero or more characters in lowers
    # - an integer is a Word composed of digits
    # - an elementRef is an element, optionally followed by an integer - if
    #   the integer is omitted, assume the value "1" as a default; these are
    #   enclosed in a Group to make it easier to walk the list of parsed
    #   chemical symbols, each with its associated number of atoms per
    #   molecule
    # - a chemicalFormula is just one or more elementRef's
    # Version 3 - Compute partial molecular weight per element, simplifying
    # summing
    # No need to redefine grammar, just define parse action function, and
    # attach to elementRef

    # Auto-convert integers, and add results names
    def convert_integers(tokens):
        return int(tokens[0])

    element = Word( caps, lowers )
    integer = Word( digits ).setParseAction( convert_integers )
    element_ref = Group( element("symbol") + Optional( integer, default=1 )("qty") )

    chemical_formula = OneOrMore( element_ref )

    def compute_element_weight(tokens):
        element = tokens[0]
        element["weight"] = atoms[element.symbol] * element.qty

    element_ref.setParseAction(compute_element_weight)

    formula_data = chemical_formula.parseString(formula)

    # compute molecular weight
    mw = sum([element.weight for element in formula_data])

    return mw

def df_to_dict_of_tuples(df, tuple_name='Row'):
    ''' changes dataframe to dictionary of tuples '''
    row_tuple = namedtuple(tuple_name, df.columns)
    tuples_dict = {}
    for row in df.itertuples():
        tuple_ = row_tuple(*row[1:])
        tuples_dict[tuple_.feature] = tuple_
    return tuples_dict


def convert_csv_to_tuples():
    ''' Convert the definitions in CSV files to one dict of named tuples and
        write that to constants.py
    '''
    default_read_csv_args = dict(na_values=[None, 'None'], comment='#')
    atoms = pd.read_csv(PATH / 'atoms.csv', **default_read_csv_args)
    ions = pd.read_csv(PATH / 'ions.csv', **default_read_csv_args)
    properties = pd.read_csv(PATH / 'other_than_concentrations.csv', **default_read_csv_args)

    # convert nan to None
    atoms = atoms.where((pd.notnull(atoms)), None)
    ions = ions.where((pd.notnull(ions)), None)
    properties = properties.where((pd.notnull(properties)), None)

    rename_columns = {'molar_weight': 'mw', 'predominant_oxidized_state': 'oxidized', 'predominant_reduced_state': 'reduced'}

    atoms = atoms.rename(columns=rename_columns)
    ions = ions.rename(columns=rename_columns)
    properties = properties.rename(columns=rename_columns)

    # remove leading and trailing white spaces
    # atoms.loc[:, ['feature', 'name', 'unit']] = atoms.loc[:, ['feature', 'name', 'unit']].applymap(lambda x: x.strip())
    # ions.loc[:, ['feature', 'name', 'unit']] = ions.loc[:, ['feature', 'name', 'unit']].applymap(lambda x: x.strip())
    # properties.loc[:, ['feature', 'name', 'unit', 'example']] = properties.loc[:, ['feature', 'name', 'unit', 'example']].applymap(lambda x: x.strip())

    # create a dict with atoms as key and mw as value to be used to parse the molecular formula in ions['feature']
    atoms_dict = atoms.set_index(['feature'])['mw'].to_dict()
    ions['mw'] = ions.apply(lambda x: _formula_parser(x['feature'], x['mw'], atoms_dict), axis=1)

    atoms_dict = df_to_dict_of_tuples(atoms, tuple_name='Atom')
    ions_dict = df_to_dict_of_tuples(ions, tuple_name='Ion')
    properties_dict = df_to_dict_of_tuples(properties, tuple_name='Properties')

    return atoms_dict, ions_dict, properties_dict